import numpy as np
from scipy import optimize
import pandas as pd
import anndata
import os
from os.path import exists
import squidpy as sq
import warnings
warnings.filterwarnings('ignore')
import time
from importlib import resources

def abstract(idata, n_neighbors=10):
    from somde import SomNode
    df = idata.to_df().T
    corinfo = idata.obs
    corinfo["total_count"]=df.sum(0)
    X=corinfo[['row','col']].values.astype(np.float32)
    som = SomNode(X,n_neighbors)
    ndf,ninfo = som.mtx(df)
    meta_idata = anndata.AnnData(ndf.T)
    meta_idata.obs[['row', 'col', 'total_count']] = ninfo.to_numpy(dtype=float)
    meta_idata.obsm['spatial'] = meta_idata.obs[['row', 'col']].to_numpy()
    idata = som_mapping(som, idata, df)
    return som, idata, meta_idata

def find_svi(idata, out_f, overwrite, som=None):
    svi_nnSVG(idata,out_f,overwrite)
    svi_scGCO(idata,out_f,overwrite)
    svi_SOMDE(idata,out_f,overwrite, som=som)
    svi_SPARKX(idata,out_f,overwrite)
    svi_SpatialDE2(idata,out_f,overwrite)
    svi_moran(idata,out_f,overwrite)
    svi_geary(idata,out_f,overwrite)

def svi_moran(idata, work_dir, overwrite=False):
    try:
        t0=time.time()
        n_perms=1000
        if (overwrite) | (not exists( f'{work_dir}moranI.csv')):
            sq.gr.spatial_neighbors(idata, key_added='spatial')
            sq.gr.spatial_autocorr(
                idata,
                mode="moran",
                n_perms=n_perms,
                n_jobs=10,
            )
            idata.uns['moranI_time'] = time.time()-t0
            idata.uns['moranI'].to_csv(f'{work_dir}moranI.csv')
        result = pd.read_csv(f'{work_dir}moranI.csv', index_col=0)
        idata.uns['moranI'] = result
        print(f'Added key moranI in idata.uns')
    except:
        pass

def svi_geary(idata, work_dir,overwrite=False):
    try:
        t0=time.time()
        n_perms=1000
        if (overwrite) | (not exists( f'{work_dir}gearyC.csv')):
            sq.gr.spatial_neighbors(idata, key_added='spatial')
            sq.gr.spatial_autocorr(
                idata,
                mode="geary",
                n_perms=n_perms,
                n_jobs=10,
            )
            idata.uns['gearyC_time'] = time.time()-t0
            idata.uns['gearyC'].to_csv(f'{work_dir}gearyC.csv')
        result = pd.read_csv(f'{work_dir}gearyC.csv', index_col=0)
        idata.uns['gearyC'] = result
        print(f'Added key gearyC in idata.uns')
    except:
        pass

def svi_nnSVG(idata, work_dir, overwrite=False):
    try:
        count_f = f'{work_dir}idata_count.csv'
        meta_f = f'{work_dir}idata_meta.csv'
        if (overwrite) | ((not exists(count_f)) & (not exists(meta_f))):
            idata.to_df().to_csv(count_f)
            idata.obs[['row', 'col']].to_csv(meta_f)
        if (overwrite) | (not exists( f'{work_dir}nnSVG.csv')):
            t0=time.time()
            with resources.path("spider.R_script", "run_nnSVG.R") as pw_fn:
                os.system(str(f'/bin/bash -c "source /etc/profile;module load GCC/11.2.0 OpenMPI/4.1.1 R/4.2.0 Anaconda3/2022.05 R-bundle-Bioconductor/3.15-R-4.2.0;R -f {pw_fn} {count_f} {meta_f} {work_dir}"'))
            idata.uns['nnSVG_time'] = time.time()-t0
        result = pd.read_csv(f'{work_dir}nnSVG.csv', index_col=0)
        idata.uns['nnSVG'] = result
        print(f'Added key nnSVG in idata.uns')
    except:
        pass
    
def svi_trendsceek(idata, work_dir, overwrite=False):
    try:
        count_f = f'{work_dir}idata_count.csv'
        meta_f = f'{work_dir}idata_meta.csv'
        if (overwrite) | ((not exists(count_f)) & (not exists(meta_f))):
            idata.to_df().to_csv(count_f)
            idata.obs[['row', 'col']].to_csv(meta_f)
        if (overwrite) | (not exists( f'{work_dir}trendsceek.csv')):
            with resources.path("spider.R_script", "run_trendsceek.R") as pw_fn:
                os.system(str(f'/bin/bash -c "source /etc/profile;module load GCC/11.2.0 OpenMPI/4.1.1 R/4.2.0 Anaconda3/2022.05 R-bundle-Bioconductor/3.15-R-4.2.0;R -f {pw_fn} {count_f} {meta_f} {work_dir}"'))
        result = pd.read_csv(f'{work_dir}trendsceek.csv', index_col=0)
        idata.uns['trendsceek'] = result
        print(f'Added key trendsceek in idata.uns')
    except:
        pass
    
def scGCO_sv(locs, data_norm, cellGraph, gmmDict, smooth_factor=10, unary_scale_factor=100, label_cost=10, algorithm='expansion'):
    from itertools import repeat
    from functools import reduce
    import operator
    import statsmodels.stats.multitest as multi
    import scGCO

    results = [scGCO.compute_single_fixed_sf(locs, data_norm, cellGraph, gmmDict, w=None, n=None, smooth_factor=10, unary_scale_factor=100, label_cost=10, algorithm='expansion')]
    
    nnn = [results[i][0] for i in np.arange(len(results))]
    nodes = reduce(operator.add, nnn)
    ppp = [results[i][1] for i in np.arange(len(results))]
    p_values=reduce(operator.add, ppp)
    ggg = [results[i][2] for i in np.arange(len(results))]
    genes = reduce(operator.add, ggg)
    # exp_ppp = [results[i][3] for i in np.arange(len(results))]
    # exp_pvalues = reduce(operator.add, exp_ppp)  
    # exp_ddd = [results[i][4] for i in np.arange(len(results))]
    # exp_diffs = reduce(operator.add, exp_ddd)      
    fff = [results[i][3] for i in np.arange(len(results))]
    s_factors = reduce(operator.add, fff)
    lll = [results[i][4] for i in np.arange(len(results))]
    pred_labels = reduce(operator.add, lll)
    mmm = [results[i][5] for i in np.arange(len(results))]
    model_labels = reduce(operator.add, mmm)

    best_p_values=[min(i) for i in p_values]
    fdr = multi.multipletests(np.array(best_p_values), method='fdr_bh')[1]
    #exp_fdr = multi.multipletests(np.array(exp_pvalues), method='fdr_bh')[1]    
    
    labels_array = np.array(pred_labels).reshape(len(genes), pred_labels[0].shape[0])
    data_array = np.array((genes, p_values, fdr,s_factors, nodes, model_labels), dtype=object).T
    t_array = np.hstack((data_array, labels_array))
    c_labels = ['p_value', 'fdr',  'smooth_factor', 'nodes','model_labels']
    for i in np.arange(labels_array.shape[1]) + 1:
        temp_label = 'label_cell_' + str(i)
        c_labels.append(temp_label)
    result_df = pd.DataFrame(t_array[:,1:], index=t_array[:,0], columns=c_labels)
    return result_df

def svi_scGCO(idata, work_dir, overwrite=False):
    try:
        import scGCO
        if (overwrite) | (not exists( f'{work_dir}scGCO.csv')):
            t0=time.time()
            data_norm = scGCO.normalize_count_cellranger(idata.to_df())
            exp= data_norm.iloc[:,0]
            pos_key=['row', 'col']
            locs = idata.obs[pos_key].to_numpy()
            cellGraph= scGCO.create_graph_with_weight(locs, exp)
            gmmDict= scGCO.gmm_model(data_norm)
            result_df= scGCO_sv(locs, data_norm, cellGraph,gmmDict)
            scGCO.write_result_to_csv(result_df, f'{work_dir}scGCO.csv')
            idata.uns['scGCO_time'] = time.time()-t0
        result = pd.read_csv(f'{work_dir}scGCO.csv')
        idata.uns['scGCO'] = result
        print(f'Added key scGCO in idata.uns')
    except:
        pass
    
def svi_SPARKX(idata, work_dir, overwrite=False):
    try:
        count_f = f'{work_dir}idata_count.csv'
        meta_f = f'{work_dir}idata_meta.csv'
        if (overwrite) | ((not exists(count_f)) & (not exists(meta_f))):
            idata.to_df().to_csv(count_f)
            idata.obs[['row', 'col']].to_csv(meta_f)
        if (overwrite) | (not exists( f'{work_dir}SPARKX.csv')):
            t0=time.time()
            with resources.path("spider.R_script", "run_SPARKX.R") as pw_fn:
                os.system(str(f'/bin/bash -c "source /etc/profile;module load GCC/11.2.0 OpenMPI/4.1.1 R/4.2.0 Anaconda3/2022.05 R-bundle-Bioconductor/3.15-R-4.2.0;R -f {pw_fn} {count_f} {meta_f} {work_dir}"'))
            idata.uns['SPARKX_time'] = time.time()-t0
        result = pd.read_csv(f'{work_dir}SPARKX.csv', index_col=0)
        idata.uns['SPARKX'] = result
        print(f'Added key SPARKX in idata.uns')
    except:
        pass
    
def svi_SpatialDE2(idata, work_dir, overwrite=False):
    try:
        # if (overwrite) | (not exists(f'{work_dir}SpatialDE.csv')) | (not exists(f'{work_dir}SpatialDE_individual.csv')):
        if (overwrite) | (not exists(f'{work_dir}SpatialDE.csv')):
            from spider import SpatialDE as SpatialDE2
            t0=time.time()
            svg_full, individual = SpatialDE2.test(idata, omnibus=False)
            svg_full = pd.concat([svg_full.set_index('gene'), individual.loc[individual.groupby('gene').lengthscale.idxmin()].set_index('gene')], axis=1)
            svg_full.to_csv(f'{work_dir}SpatialDE.csv')
            individual.to_csv(f'{work_dir}SpatialDE_individual.csv')
            idata.uns['SpatialDE2_time'] = time.time()-t0
        result = pd.read_csv(f'{work_dir}SpatialDE.csv', index_col=0)
        idata.uns['SpatialDE'] = result
        print(f'Added key SpatialDE in idata.uns')
    except:
        pass
    
    
def svi_SOMDE(idata, work_dir, overwrite=False, som=None):
    try:
        if (overwrite) | (not exists(f'{work_dir}SOMDE.csv')):
            t0=time.time()
            if som is None:
                from somde import SomNode
                df = idata.to_df().T
                corinfo = idata.obs
                corinfo["total_count"]=df.sum(0)
                X=corinfo[['row','col']].values.astype(np.float32)
                som = SomNode(X,10)
                ndf,ninfo = som.mtx(df)
            nres = som.norm()
            result, SVnum =som.run()
            result.to_csv(f'{work_dir}SOMDE.csv')
            idata.uns['SOMDE_time'] = time.time()-t0
        result = pd.read_csv(f'{work_dir}SOMDE.csv', index_col=0)
        idata.uns['SOMDE'] = result
        print(f'Added key SOMDE in idata.uns')
    except:
        pass
    
def combine_SVI(idata, threshold):
    svi_df, svi_df_strict = combine_SVI_strict(idata,threshold=threshold)
    if len(svi_df_strict) < 10:
        print('Detected SVI number is less than 10, falling back to relaxed filtering.')
        svi_df, svi_df_strict = combine_SVI_Fisher(idata,threshold=threshold)
    if len(svi_df_strict) < 10:
        print('Detected SVI number is less than 10, falling back to use SOMDE result only.')
        svi_df, svi_df_strict  = combine_SVI_somde(idata,threshold=threshold)
    return svi_df, svi_df_strict

def combine_SVI_Fisher(idata, threshold=0.05):
    from scipy.stats import combine_pvalues
    methods = np.array(['SOMDE', 'SpatialDE', 'SPARKX', 'nnSVG', 'scGCO', 'gearyC', 'moranI'])[np.isin(['SOMDE', 'SpatialDE', 'SPARKX', 'nnSVG', 'scGCO', 'gearyC', 'moranI'],list(idata.uns.keys()))]
    df = []
    for i in methods:
        if i == 'SOMDE':
            df.append(idata.uns[i].set_index('g')[['pval']].rename(columns = {'pval': i}))
        elif i == 'SpatialDE':
            df.append(idata.uns[i][['padj']].rename(columns = {'padj': i}))
            # df.append(idata.uns[i].set_index('gene')[['padj']].rename(columns = {'padj': i}))
        elif i == 'SPARKX':
            df.append(idata.uns[i][['adjustedPval']].rename(columns = {'adjustedPval': i}))
        elif i == 'nnSVG':
            df.append(idata.uns[i][['padj']].rename(columns = {'padj': i}))
        elif i == 'scGCO':
            df.append(idata.uns[i].set_index(idata.uns['scGCO'].columns[0])[['fdr']].rename(columns = {'fdr': i}))
    df = pd.concat(df, axis=1).fillna(1)
    
    arr = [combine_pvalues(x, method='fisher')[1] for x in df.to_numpy()]
    df['adj_pvalue'] = arr
    df_sub = df[df['adj_pvalue']<threshold]
    # df_sub = df_sub.loc[np.intersect1d(df_sub.index, idata.uns['SOMDE']['g'])]
    print(f'{len(df_sub)}/{len(df)} SVIs identified (threshold={threshold}).')
    idata.varm['svi'] = df.loc[idata.var_names]
    idata.var['is_svi'] = 0
    idata.var.loc[df_sub.index, 'is_svi'] = 1
    return df, df_sub

def combine_SVI_strict(idata, threshold=0.01):
    methods = np.array(['SOMDE', 'SpatialDE', 'SPARKX', 'nnSVG', 'scGCO', 'gearyC', 'moranI'])[np.isin(['SOMDE', 'SpatialDE', 'SPARKX', 'nnSVG', 'scGCO', 'gearyC', 'moranI'],list(idata.uns.keys()))]
    print(f'Using the results from SVI identification methods: {methods}')
    df = []
    for i in methods:
        if i == 'SOMDE':
            df.append(idata.uns[i].set_index('g')[['pval']].rename(columns = {'pval': i}))
        elif i == 'SpatialDE':
            df.append(idata.uns[i][['padj']].rename(columns = {'padj': i}))
        elif i == 'SPARKX':
            df.append(idata.uns[i][['adjustedPval']].rename(columns = {'adjustedPval': i}))
        elif i == 'nnSVG':
            df.append(idata.uns[i][['padj']].rename(columns = {'padj': i}))
        elif i == 'scGCO':
            df.append(idata.uns[i].set_index(idata.uns['scGCO'].columns[0])[['fdr']].rename(columns = {'fdr': i}))
    df = pd.concat(df, axis=1).fillna(1)
    df_sub = df[(df<threshold).all(axis=1)]
    print(f'{len(df_sub)}/{len(df)} SVIs identified (threshold={threshold}).')
    idata.varm['svi'] = df.loc[idata.var_names]
    idata.var['is_svi'] = 0
    idata.var.loc[df_sub.index, 'is_svi'] = 1
    return df, df_sub

def combine_SVI_somde(idata, threshold=0.01):
    print(f'Using the SOMDE results')
    df = idata.uns['SOMDE'].set_index('g')[['pval']].rename(columns = {'pval': 'SOMDE'}).fillna(1)
    df_sub = df[(df<threshold).all(axis=1)]
    print(f'{len(df_sub)}/{len(df)} SVIs identified (threshold={threshold}).')
    idata.varm['svi'] = df.loc[idata.var_names]
    idata.var['is_svi'] = 0
    idata.var.loc[df_sub.index, 'is_svi'] = 1
    return df, df_sub
    

def eva_SVI(idata, svi_df_strict):
    import seaborn as sns
    dfs = [
        -idata.uns['gearyC'][['C']],
        idata.uns['moranI'][['I']],
        idata.uns['SOMDE'].set_index('g')['FSV'],
        idata.uns['nnSVG']['LR_stat'],
        idata.uns['SOMDE'].set_index('g')['LLR'],
    ]
    df = pd.concat(dfs, axis=1)
    metrics = ["Geary C (rev.)", "Moran I", 'FSV', 'LR (nnSVG)', 'LLR (SOMDE)',]
    df.columns=metrics

    normalized_df=(df-df.min())/(df.max()-df.min())
    normalized_df['Category'] = 'Excluded'
    normalized_df.loc[svi_df_strict.index, 'Category'] = 'SVI'
    normalized_df = normalized_df.melt(id_vars='Category', value_vars=metrics, var_name='Metric')

    ax =sns.boxplot(normalized_df,x='Metric',y='value', hue='Category', palette={'SVI':'#80b1d3','Excluded': '#fb8072'}, width=0.8, hue_order=['SVI', 'Excluded'])
    ax.legend(loc='upper center',ncol=2, bbox_to_anchor=(0.5, 1.1), frameon=False)
    ax.set_ylabel('')    
    ax.set_xlabel('')
    
class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

def SVI_patterns(idata, svi_df_strict, iter=1000, pattern_prune_threshold=1e-8):
    from spider import SpatialDE as SpatialDE2
    allsignifgenes = svi_df_strict.index.to_numpy()
    if 'lengthscale' in idata.uns['SpatialDE'].columns:
        l=idata.uns['SpatialDE'].loc[allsignifgenes]['lengthscale'].to_list()
        if len(np.unique(l)) < 2:
            l=idata.uns['SOMDE'].set_index('g').loc[allsignifgenes]['l'].to_list()
    else:
        l=idata.uns['SOMDE'].set_index('g').loc[allsignifgenes]['l'].to_list()
    pattern_number = 1000
    if len(np.unique(l)) <= 1:
        pattern_number = -1
    for count in range(5):
        print(pattern_number, count, pattern_prune_threshold)
        if (pattern_number > 100) and (pattern_prune_threshold<1) and (count < 4):
            param_obj = SpatialDE2.SpatialPatternParameters(lengthscales=l,maxiter=iter, pattern_prune_threshold=pattern_prune_threshold)
            upper_patterns, _ = SpatialDE2.spatial_patterns(idata, genes=allsignifgenes, rng=np.random.default_rng(seed=45), params=param_obj, copy=False)
            pattern_number = upper_patterns.patterns.shape[1]
            pattern_prune_threshold = pattern_prune_threshold*100
            if pattern_number < 2:
                pattern_number = -1
        elif (pattern_number < 100) and (pattern_number > 2):
            break
        else:
            print(f'falling back to controlled pattern')
            histology_results, patterns, prob = SVI_patterns_v1(idata, svi_df_strict)
            upper_patterns = dotdict({
                'labels': histology_results['pattern'].to_numpy(),
                'patterns': patterns.to_numpy(),
                'pattern_probabilities': prob
            })
            pattern_number = patterns.shape[1]
            break

    print(f'eventually found {pattern_number} patterns')
    idata.var['label'] = -1
    idata.var.loc[allsignifgenes, 'label'] = upper_patterns.labels
    idata.var[[f'pattern_membership_{x}' for x in range(upper_patterns.pattern_probabilities.shape[1])]] = 0
    idata.var.loc[allsignifgenes, [f'pattern_membership_{x}' for x in range(upper_patterns.pattern_probabilities.shape[1])]] = upper_patterns.pattern_probabilities
    idata.obsm['pattern_score'] = upper_patterns.patterns
    
def SVI_patterns_v1(idata, svi_df_strict, components=5):
    import NaiveDE
    import SpatialDE
    df = idata.to_df().T.loc[svi_df_strict.index]
    corinfo = idata.obs
    corinfo["total_counts"]=df.sum(0)
    X=corinfo[['row','col']].values.astype(np.float32)
    norm_expr = NaiveDE.stabilize(df).T
    resid_expr = NaiveDE.regress_out(corinfo, norm_expr.T, 'np.log(total_counts)').T
    print('finished regression')
    results = SpatialDE.run(X,resid_expr)
    print('finished fitting')
    histology_results, patterns, prob = spatial_patterns(X, resid_expr, results, C=components,l=results['l'].median()+0.5, verbosity=1)
    return histology_results, patterns, prob

def som_mapping(som, idata, df):
    bsmc = som.som.bmus
    soml = []
    for i in np.arange(bsmc.shape[0]):
        u,v = bsmc[i]
        soml.append(v*som.somn+u)
    idata.obs['som_node'] = -1
    ids = np.sort(np.unique(np.array(soml)))
    count = 0
    for i in ids:
        idata.obs.loc[df.loc[:,np.array(soml)==i].columns,'som_node'] = count
        count += 1
    return idata

def meta_pattern_to_idata(idata, meta_idata):
    idata.obsm['pattern_score'] = meta_idata.obsm['pattern_score'][idata.obs['som_node'].to_numpy()]  
    print(f'Added key pattern_score in idata.obsm')     
    
def idata_pattern_to_spot(idata):
    belonging = {}
    cells = idata.uns['cell_meta'].index
    for i in cells:
        belonging[i] = []
    for pair in idata.obs.reset_index()[['index','A', 'B']].to_numpy():
        belonging[pair[1]].append(pair[0])
        belonging[pair[2]].append(pair[0])
    score = pd.DataFrame(idata.obsm['pattern_score'], index=idata.obs_names)
    df = pd.concat([score.loc[belonging[c]].mean() for c in cells], axis=1).T     
    df.index = cells
    idata.uns['cell_pattern'] = df
    print(f'Added key cell_pattern in idata.uns')   
    return df

def pattern_label_corr(data, pattern_df, label_key):
    from sklearn.feature_selection import mutual_info_classif
    label_df = pd.get_dummies(data.obs[label_key]).T
    mi=pd.DataFrame([mutual_info_classif(pattern_df, x) for x in label_df.to_numpy()], index=label_df.index, columns=pattern_df.columns)
    corr = pd.concat([pattern_df, label_df.T], axis=1).corr().loc[label_df.index,  pattern_df.columns]
    return mi, corr

def spatial_patterns(X, exp_mat, DE_mll_results, C, l, **kwargs):
    ''' Group spatially variable genes into spatial patterns using
    automatic expression histology (AEH).
    X : Spatial coordinates
    exp_mat : Expression matrix, appropriately normalised.
    DE_mll_results : Results table from SpatialDE, after filtering
        for significance level.
    C : The number of spatial patterns
    **kwards are passed on to the function fit_patterns()
    Returns
    pattern_results : A DataFrame with pattern membership information
        for each gene
    patterns : The posterior mean underlying expression for genes in
        given spatial patterns.
    '''
    Y = exp_mat[DE_mll_results['g']].values.T

    # This is important, we only care about co-expression, not absolute levels.
    Y = (Y.T - Y.mean(1)).T
    Y = (Y.T / Y.std(1)).T

    _, m, r, _ = fit_patterns(X, Y, C, l, **kwargs)

    cres = pd.DataFrame({'g': DE_mll_results['g'],
                         'pattern': r.argmax(1),
                         'membership': r.max(1)})

    m = pd.DataFrame.from_records(m)
    m.index = exp_mat.index

    return cres, m, r


def make_elbojective(Y, r, m, X, K_0, s2e_0, pi=None):
    def elbojective(log_s2e):
        return -ELBO(Y, r, m, np.exp(log_s2e), K_0, K_0, s2e_0, pi)
    
    return elbojective

def SE_kernel(X, l):
    X = np.array(X)
    Xsq = np.sum(np.square(X), 1)
    R2 = -2. * np.dot(X, X.T) + (Xsq[:, None] + Xsq[None, :])
    R2 = np.clip(R2, 1e-12, np.inf)
    return np.exp((-R2 / (2 * l ** 2)).astype(float))

def ELBO(Y, r, m, s2e, K, K_0, s2e_0, pi=None):
    L = ln_P_YZms(Y, r, m, s2e, pi) + ln_P_Z(r, pi) + ln_P_mu(m, K) \
        - ln_Q_Z(r, r) - ln_Q_mu(K_0, r, s2e_0)
    
    return L


def factor(K):
    S, U = np.linalg.eigh(K)
    # .clip removes negative eigenvalues
    return U, np.clip(S, 1e-8, None)

def ln_Q_mu(K, Z, s2e):
    ''' Expecation of ln Q(mu)
    '''
    N = K.shape[0]
    C = Z.shape[1]
    G_k = Z.sum(0)
    
    ll = 0
    U, S = factor(K)
    for k in range(C):
        ll = ll - (1. / S + G_k[k] / s2e).sum()
        ll = ll + N * np.log(2 * np.pi)
        
    
    ll = -0.5 * ll
    
    return ll

def ln_Q_Z(Z, r):
    ''' Expectation of ln Q(Z)
    '''
    return np.sum(Z * np.log(r))

def ln_P_YZms(Y, Z, mu, s2e, pi=None):
    ''' Expecation of ln P(Y | Z, mu, s2e)
    '''
    G = Y.shape[0]
    N = Y.shape[1]
    C = Z.shape[1]
    if pi is None:
        pi = np.ones(C) / C
    
    log_rho = np.log(pi[None, :]) \
              - 0.5 * N * np.log(s2e) \
              - 0.5 * np.sum((mu.T[None, :, :] - Y[:, None, :]) ** 2, 2) / s2e \
              - 0.5 * N * np.log(2 * np.pi)
            
    return (Z * log_rho).sum()

def ln_P_Z(Z, pi=None):
    ''' Expectation of ln P(Z)
    '''
    C = Z.shape[1]
    if pi is None:
        pi = np.ones(C) / C
        
    return np.dot(Z, np.log(pi)).sum()

def ln_P_mu(mu, K):
    ''' Expectation of ln P(mu)
    '''
    N = K.shape[0]
    C = mu.shape[1]
    ll = 0
    for k in range(C):
        ll = ll + np.linalg.det(K)
        ll = ll + mu[:, k].dot(np.linalg.solve(K, mu[:, k]))
        ll = ll + N * np.log(2 * np.pi)
        
    ll = -0.5 * ll
    
    return ll

def fit_patterns(X, Y, C, l, s2e_0=1.0, verbosity=0, maxiter=100, printerval=1, opt_interval=1, delta_elbo_threshold=1e-2):
    ''' Fit spatial patterns using Automatic Expression Histology.
    X : Spatial coordinates
    Y : Gene expression values
    C : The number of patterns
    l : The charancteristic length scale of the clusters
    Returns
    final_elbo : The final ELBO value.
    m : The posterior mean underlying expression values for each cluster.
    r : The posterior pattern assignment probabilities for each gene and pattern.
    s2e : The estimated noise parameter of the model
    '''
    # Set up constants
    G = Y.shape[0]
    N = Y.shape[1]
    eps = 1e-8 * np.eye(N)
    
    s2e = s2e_0
    
    K = SE_kernel(X, l) + eps
    
    # Randomly initialize
    r = np.random.uniform(size=(G, C))
    r = r / r.sum(0)
    
    pi = r.sum(0) / G

    m = np.random.normal(size=(N, C))
    
    elbo_0 = ELBO(Y, r, m, s2e, K, K, s2e, pi)
    elbo_1 = elbo_0

    if verbosity > 0:
        print('iter {}, ELBO: {:0.2e}'.format(0, elbo_1))

    if verbosity > 1:
        print()

    for i in range(maxiter):
        if (i % opt_interval == (opt_interval - 1)):
            elbojective = make_elbojective(Y, r, m, X, K, s2e, pi)
            
            o = optimize.minimize_scalar(elbojective)
            s2e = np.exp(o.x)
            
            
        r = Q_Z_expectation(m, Y, s2e, N, C, G, pi)
        m = Q_mu_expectation(r, Y, K, s2e)
        
        pi = r.sum(0) / G

        elbo_0 = elbo_1
        elbo_1 = ELBO(Y, r, m, s2e, K, K, s2e, pi)
        delta_elbo = np.abs(elbo_1 - elbo_0)

        if verbosity > 0 and (i % printerval == 0):
            print('iter {}, ELBO: {:0.2e}, delta_ELBO: {:0.2e}'.format(i + 1, elbo_1, delta_elbo))
            
            if verbosity > 1:
                print('ln(l): {:0.2f}, ln(s2e): {:.2f}'.format(np.log(l), np.log(s2e)))
                
            if verbosity > 2:
                line1 = 'P(Y | Z, mu, s2e): {:0.2e}, P(Z): {:0.2e}, P(mu): {:0.2e}' \
                        .format(ln_P_YZms(Y, r, m, s2e, pi), ln_P_Z(r, pi), ln_P_mu(m, K))
                line2 = 'Q(Z): {:0.2e}, Q(mu): {:0.2e}'.format(ln_Q_Z(r, r), ln_Q_mu(K, r, s2e))
                print(line1 + '\n' + line2)
            
            if verbosity > 1:
                print()
            
        if delta_elbo < delta_elbo_threshold:
            if verbosity > 0:
                print('Converged on iter {}'.format(i + 1))

            break
            
    else:
        print('Warning! ELBO dit not converge after {} iters!'.format(i + 1))

    final_elbo = ELBO(Y, r, m, s2e, K, K, s2e, pi)
        
    return final_elbo, m, r, s2e

def Q_Z_expectation(mu, Y, s2e, N, C, G, pi=None):
    if pi is None:
        pi = np.ones(C) / C

    log_rho = np.log(pi[None, :]) \
              - 0.5 * N * np.log(s2e) \
              - 0.5 * np.sum((mu.T[None, :, :] - Y[:, None, :]) ** 2, 2) / s2e \
              - 0.5 * N * np.log(2 * np.pi)

    # Subtract max per row for numerical stability, and add offset from 0 for same reason.
    rho = np.exp(log_rho - log_rho.max(1)[:, None]) + 1e-12
    # Then evaluate softmax
    r = (rho.T / (rho.sum(1))).T
    
    return r

def Q_mu_k_expectation(Z_k, Y, K, s2e):
    y_k_tilde = np.dot(Z_k, Y) / s2e
    Sytk = np.dot(K, y_k_tilde)
    IpSDk = np.eye(K.shape[0]) + K * Z_k.sum() / s2e
    m_k = np.linalg.solve(IpSDk, Sytk)
    
    return m_k


def Q_mu_expectation(Z, Y, K, s2e):
    m = np.zeros((Y.shape[1], Z.shape[1]))

    y_k_tilde = np.dot(Z.T, Y).T / s2e

    for k in range(Z.shape[1]):
        m[:, k] = Q_mu_k_expectation(Z[:, k], Y, K, s2e)

    return m
    


import logging
from time import time
import warnings
from typing import Optional, Dict, List
from enum import Enum, auto

import numpy as np
import pandas as pd

from tqdm.auto import tqdm

import tensorflow as tf

import NaiveDE
from anndata import AnnData

from ._internal.kernels import SquaredExponential, Cosine, Linear
from ._internal.models import Model, Constant, Null, model_factory
from ._internal.util import (
    DistanceCache,
    default_kernel_space,
    kspace_walk,
    dense_slice,
    normalize_counts,
)
from ._internal.tf_dataset import AnnDataDataset
from ._internal.gpflow_helpers import *


class GP(Enum):
    GPR = auto()
    """
    Dense Gaussian process.
    """
    SGPR = auto()
    """
    Sparse Gaussian process.
    """


class SGPIPM(Enum):
    free = auto()
    """Inducing points are initialized randomly and their positions are optimized together with the other parameters."""
    random = auto()
    """Inducing points are placed at random locations."""
    grid = auto()
    """Inducing points are placed in a regular grid."""


@dataclass(frozen=True)
class GPControl:
    """
    Parameters for Gaussian process fitting.

    Args:
        gp: Type of GP to fit.
        ipm: Inducing point method. Only used if ``gp == GP.SGPR``.
        ncomponents: Number of kernel components.
        ard: Whether to use automatic relevance determination. This amounts to having one
            lengthscale per spatial dimension.
        ninducers: Number of inducing points.
    """

    gp: Optional[GP] = None
    ipm: SGPIPM = SGPIPM.grid
    ncomponents: int = 5
    ard: bool = True
    ninducers: Optional[int] = None


def inducers_grid(X, ninducers):
    rngmin = X.min(0)
    rngmax = X.max(0)
    xvals = np.linspace(rngmin[0], rngmax[0], int(np.ceil(np.sqrt(ninducers))))
    yvals = np.linspace(rngmin[1], rngmax[1], int(np.ceil(np.sqrt(ninducers))))
    xx, xy = np.meshgrid(xvals, yvals)
    return np.hstack((xx.reshape((xx.size, 1)), xy.reshape((xy.size, 1))))


def fit_model(model: Model, genes: Union[List[str], np.ndarray], counts: np.ndarray):
    results = []
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        with model:
            for i, gene in enumerate(tqdm(genes)):
                y = dense_slice(counts[:, i])
                model.y = y
                t0 = time()

                res = model.optimize()
                t = time() - t0
                res = {
                    "gene": gene,
                    "max_ll": model.log_marginal_likelihood,
                    "max_delta": model.delta,
                    "max_mu_hat": model.mu,
                    "max_s2_t_hat": model.sigma_s2,
                    "max_s2_e_hat": model.sigma_n2,
                    "time": t,
                    "n": model.n,
                    "FSV": model.FSV,
                    "s2_FSV": np.abs(
                        model.s2_FSV
                    ),  # we are at the optimum, so this should never be negative.
                    "s2_logdelta": np.abs(
                        model.s2_logdelta
                    ),  # Negative results are due to numerical errors when evaluating vanishing Hessians
                    "converged": res.success,
                    "M": model.n_parameters,
                }
                for (k, v) in vars(model.kernel).items():
                    if k not in res and k[0] != "_":
                        res[k] = v

                results.append(res)
    return pd.DataFrame(results)


def fit_detailed(
    adata: AnnData,
    genes: Optional[List[str]] = None,
    layer: Optional[str] = None,
    normalized: bool = False,
    sizefactor_col: Optional[str] = None,
    spatial_key: str = "spatial",
    control: Optional[GPControl] = GPControl(),
    rng: np.random.Generator = np.random.default_rng(),
) -> DataSetResults:
    """
    Fits Gaussian processes to genes.

    A Gaussian process based on highly interpretable spectral mixture kernels (Wilson et al. 2013, Wilson 2014) is fitted
    separately to each gene. Sparse GPs are used on large datasets (>1000 observations) to improve speed.
    This uses a Gaussian likelihood and requires appropriate data normalization.

    Args:
        adata: The annotated data matrix.
        genes: List of genes to base the analysis on. Defaults to all genes.
        layer: Name of the AnnData object layer to use. By default ``adata.X`` is used.
        normalized: Whether the data are already normalized to an approximately Gaussian likelihood.
            If ``False``, they will be normalized using the workflow from Svensson et al, 2018.
        sizefactor_col: Column in ``adata.obs`` to be used for normalization. If ``None``, total number of
            counts per spot will be used.
        spatial_key: Key in ``adata.obsm`` where the spatial coordinates are stored.
        control: Parameters for the Gaussian process, e.g. number of kernel components, number of inducing points.
        rng: Random number generator.

    Returns:
        A dictionary with the results. The dictionary has an additional method ``to_df``, which returns a DataFrame
        with the key results.
    """
    if not normalized and genes is None:
        warnings.warn(
            "normalized is False and no genes are given. Assuming that adata contains complete data set, will normalize and fit a GP for every gene."
        )

    if not normalized:
        adata = normalize_counts(adata, sizefactor_col, layer, copy=True)

    data = adata[:, genes] if genes is not None else adata
    X = data.obsm[spatial_key]
    counts = data.X if layer is None else data.layers[layer]

    gp = control.gp
    if gp is None:
        if data.n_obs < 1000:
            gp = GP.GPR
        else:
            gp = GP.SGPR

    results = DataSetResults()
    X = tf.convert_to_tensor(X, dtype=gpflow.config.default_float())
    t = tqdm(data.var_names)
    opt = gpflow.optimizers.Scipy()

    logging.info("Fitting gene models")
    if gp == GP.GPR:
        for g, gene in enumerate(t):
            t.set_description(gene, refresh=False)
            model = GPR(
                X,
                Y=tf.convert_to_tensor(
                    dense_slice(counts[:, g])[:, np.newaxis],
                    dtype=gpflow.config.default_float(),
                ),
                n_kernel_components=control.ncomponents,
                ard=control.ard,
            )
            results[gene] = GeneGP(model, opt.minimize, method="bfgs")
    elif gp == GP.SGPR:
        ninducers = (
            np.ceil(np.sqrt(data.n_obs)).astype(np.int32)
            if control.ninducers is None
            else control.ninducers
        )
        if control.ipm == SGPIPM.free or control.ipm == SGPIPM.random:
            inducers = X[rng.integers(0, X.shape[0], ninducers), :]
        elif control.ipm == SGPIPM.grid:
            rngmin = tf.reduce_min(X, axis=0)
            rngmax = tf.reduce_max(X, axis=0)
            xvals = tf.linspace(rngmin[0], rngmax[0], int(np.ceil(np.sqrt(ninducers))))
            yvals = tf.linspace(rngmin[1], rngmax[1], int(np.ceil(np.sqrt(ninducers))))
            xx, xy = tf.meshgrid(xvals, yvals)
            inducers = tf.stack((tf.reshape(xx, (-1,)), tf.reshape(xy, (-1,))), axis=1)
            inducers = gpflow.inducing_variables.InducingPoints(inducers)
        if control.ipm != SGPIPM.free:
            gpflow.utilities.set_trainable(inducers, False)

        method = "BFGS"
        if control.ipm == SGPIPM.free and ninducers > 1e3:
            method = "L-BFGS-B"

        for g, gene in enumerate(t):
            t.set_description(gene, refresh=False)
            model = SGPR(
                X,
                Y=tf.constant(
                    dense_slice(counts[:, g])[:, np.newaxis],
                    dtype=gpflow.config.default_float(),
                ),
                inducing_variable=inducers,
                n_kernel_components=control.ncomponents,
                ard=control.ard,
            )
            results[gene] = GeneGP(model, opt.minimize, method=method)

    logging.info("Finished fitting models to %i genes" % data.n_vars)
    return results


def fit_fast(
    adata: AnnData,
    genes: Optional[List[str]] = None,
    layer: Optional[str] = None,
    normalized: bool = False,
    sizefactor_col: Optional[str] = None,
    sparse: Optional[bool] = None,
    spatial_key: str = "spatial",
    kernel_space: Optional[Dict[str, Union[float, List[float]]]] = None,
) -> pd.DataFrame:
    """
    Fits Gaussian processes to genes.

    This uses the inflexible but fast technique of Svensson et al. (2018). In particular, the kernel lengthscale is not
    optimized, but must be given beforehand. Multiple kernel functions and lengthscales can be specified, the best-fitting
    model will be retained. To further improve speed, sparse GPs are used for large (>1000 observations) data sets with
    inducing points located on a regular grid.

    Args:
        adata: The annotated data matrix.
        genes: List of genes to base the analysis on. Defaults to all genes.
        layer: Name of the AnnData object layer to use. By default ``adata.X`` is used.
        normalized: Whether the data are already normalized to an approximately Gaussian likelihood.
            If ``False``, they will be normalized using the workflow from Svensson et al, 2018.
        sizefactor_col: Column in ``adata.obs`` to be used for normalization. If ``None``, total number of
            counts per spot will be used.
        spatial_key: Key in ``adata.obsm`` where the spatial coordinates are stored.
        sparse: Whether to use sparse GPs. Slightly faster on large datasets, but less precise.
            Defaults to sparse GPs if more than 1000 data points are given.
        kernel_space: Kernels to test against. Dictionary with the name of the kernel function as key and list of
            lengthscales (if applicable) as values. Currently, three kernel functions are known:

            * ``SE``, the squared exponential kernel :math:`k(\\boldsymbol{x}^{(1)}, \\boldsymbol{x}^{(2)}; l) = \\exp\\left(-\\frac{\\lVert \\boldsymbol{x}^{(1)} - \\boldsymbol{x}^{(2)} \\rVert}{l^2}\\right)`
            * ``PER``, the periodic kernel :math:`k(\\boldsymbol{x}^{(1)}, \\boldsymbol{x}^{(2)}; l) = \\cos\\left(2 \pi \\frac{\\sum_i (x^{(1)}_i - x^{(2)}_i)}{l}\\right)`
            * ``linear``, the linear kernel :math:`k(\\boldsymbol{x}^{(1)}, \\boldsymbol{x}^{(2)}) = (\\boldsymbol{x}^{(1)})^\\top \\boldsymbol{x}^{(2)}`

            By default, 5 squared exponential and 5 periodic kernels with lengthscales spanning the range of the
            data will be used.

    Returns:
        A Pandas DataFrame with the results.
    """
    if not normalized and genes is None:
        warnings.warn(
            "normalized is False and no genes are given. Assuming that adata contains complete data set, will normalize and fit a GP for every gene."
        )

    if not normalized:
        adata = normalize_counts(adata, sizefactor_col, layer, copy=True)

    data = adata[:, genes] if genes is not None else adata

    X = data.obsm[spatial_key]
    counts = data.X if layer is None else data.layers[layer]

    dcache = DistanceCache(X)
    if kernel_space is None:
        kernel_space = default_kernel_space(dcache)

    logging.info("Fitting gene models")
    n_models = 0
    Z = None
    if sparse is None and X.shape[0] > 1000 or sparse:
        Z = inducers_grid(X, np.maximum(100, np.sqrt(data.n_obs)))

    results = []
    for kern, kname in kspace_walk(kernel_space, dcache):
        model = model_factory(X, Z, kern)
        res = fit_model(model, data.var_names, counts)
        res["model"] = kname
        results.append(res)
        n_models += 1

    n_genes = data.n_vars
    logging.info("Finished fitting {} models to {} genes".format(n_models, n_genes))

    results = pd.concat(results, sort=True).reset_index(drop=True)
    sizes = (
        results.groupby(["model", "gene"], sort=False).size().groupby("model", sort=False).unique()
    )
    results = results.set_index("model")
    results.loc[sizes > 1, "M"] += 1
    results = results.reset_index()
    results["BIC"] = -2 * results["max_ll"] + results["M"] * np.log(results["n"])

    results = results.loc[results.groupby(["model", "gene"], sort=False)["max_ll"].idxmax()]
    results = results.loc[results.groupby("gene", sort=False)["BIC"].idxmin()]

    return results.reset_index(drop=True)


def fit(
    adata: AnnData,
    genes: Optional[List[str]] = None,
    layer: Optional[str] = None,
    normalized=False,
    spatial_key: str = "spatial",
    control: Optional[GPControl] = GPControl(),
    kernel_space: Optional[Dict[str, float]] = None,
    rng: np.random.Generator = np.random.default_rng(),
) -> pd.DataFrame:
    """
    Fits Gaussian processes to genes.

    This dispatches to :py:func:`fit_fast` if ``control`` is ``None``, otherwise to :py:func:`fit_detailed`.
    All arguments are forwarded.

    Returns: A Pandas DataFrame with the results.
    """
    if control is None:
        return fit_fast(adata, genes, layer, normalized, spatial_key, kernel_space)
    else:
        return (
            fit_detailed(adata, genes, layer, normalized, spatial_key, control, rng)
            .to_df(modelcol="model")
            .reset_index(drop=True)
        )



import pandas as pd
from os.path import exists

from . import svi
from . import preprocess
from . import clustering
from . import enrichment
from . import visualization
from . import util
from . import trajectory

class SPIDER():
    def __init__(self):
        self.svi = svi
        self.pp = preprocess
        self.cl = clustering
        self.er = enrichment
        self.vis = visualization
        self.util = util
        self.traj = trajectory
        pass


    def prep(self,
            adata_input, work_dir, 
            no_spatalk=False,
            cluster_key='type', 
            # pos_key=['row', 'col'], 
            is_human=True, 
            n_neighs=6, 
            coord_type='grid', 
            imputation=True,
            overwrite=False,
    ):
        adata = adata_input.copy()
        del adata_input
        # detect cci
        if not no_spatalk:
            out_f = f'{work_dir}/spatalk'
            if overwrite | (not exists(f'{out_f}_lrpair.csv')):
                preprocess.cci_spatalk(adata, work_dir, cluster_key, is_human, out_f)
            else:
                print(f'{out_f}_lrpair.csv already exists, skipping spatalk.')
            lr_raw = pd.read_csv(f'{out_f}_lrpair.csv', index_col=0)
        else:
            lr_raw = preprocess.load_lr_df(is_human)     
        if imputation:
            'Running imputation with MAGIC'
            preprocess.impute_MAGIC(adata)

        # idata generation
        lr_df, adata = preprocess.subset(adata, lr_raw[['ligand', 'receptor']].drop_duplicates())
        pairs = preprocess.find_pairs(adata, coord_type, n_neighs)
        pairs_meta = preprocess.meta(adata, cluster_key, pairs)
        score = preprocess.score(adata, lr_df, pairs)
        idata = preprocess.idata_construct(score, pairs_meta, lr_df, lr_raw, adata)
        return idata

    def find_svi(self, idata, out_f, abstract=True, overwrite=False, n_neighbors=10, threshold=0.01, pattern_prune_threshold=0.0001):
        from os.path import exists
        from os import mkdir
        if not exists(out_f):
            print(f'Creating folder {out_f}')
            mkdir(out_f)
        if abstract:
            som, idata, meta_idata = svi.abstract(idata, n_neighbors)
            svi.find_svi(meta_idata,out_f,overwrite, som=som) #generating results
            print('finished running all SVI tests')
            svi_df, svi_df_strict = svi.combine_SVI(meta_idata,threshold=threshold)
            if (overwrite) | (not exists(f'{out_f}pattern.csv')):
                svi.SVI_patterns(meta_idata, svi_df_strict, pattern_prune_threshold=pattern_prune_threshold)
                pd.DataFrame(meta_idata.obsm['pattern_score']).to_csv(f'{out_f}pattern.csv')
                meta_idata.var.to_csv(f'{out_f}membership.csv')
            else:
                meta_idata.obsm['pattern_score'] = pd.read_csv(f'{out_f}pattern.csv', index_col=0).to_numpy()
                meta_idata.var = pd.read_csv(f'{out_f}membership.csv', index_col=0)
            svi.meta_pattern_to_idata(idata, meta_idata)
            pd.DataFrame(meta_idata.obsm['pattern_score']).to_csv(f'{out_f}full_pattern.csv')
            return idata, meta_idata
        else:
            svi.find_svi(idata, out_f,overwrite) #generating results
            svi_df, svi_df_strict = svi.combine_SVI(idata,threshold=threshold)
            # if len(svi_df_strict) < 10:
            #     svi_df, svi_df_strict = svi.combine_SVI_Fisher(idata,threshold=threshold)
            #     print('Detected SVI number is less than 10, falling back to relaxed filtering.')
            # if len(svi_df_strict) < 10:
            #     svi_df, svi_df_strict  = svi.combine_SVI_somde(idata,threshold=threshold)
            #     print('Detected SVI number is less than 10, falling back to use SOMDE result only.')
            if (overwrite) | (not exists(f'{out_f}pattern.csv')):
                svi.SVI_patterns(idata, svi_df_strict, pattern_prune_threshold=pattern_prune_threshold)
                pd.DataFrame(idata.obsm['pattern_score']).to_csv(f'{out_f}pattern.csv')
                idata.var.to_csv(f'{out_f}membership.csv')
            else:
                idata.obsm['pattern_score'] = pd.read_csv(f'{out_f}pattern.csv', index_col=0)
                idata.var = pd.read_csv(f'{out_f}membership.csv', index_col=0)   
        idata.var[[f'pattern_correlation_{x}' for x in range(idata.obsm['pattern_score'].shape[1])]] = 0
        corr_df=pd.concat([idata[:,idata.var['is_svi']==1].to_df(),pd.DataFrame(idata.obsm['pattern_score'],index=idata.obs_names)],axis=1).corr().loc[idata[:,idata.var['is_svi']==1].var_names, range(idata.obsm['pattern_score'].shape[1])]
        idata.var.loc[idata[:,idata.var['is_svi']==1].var_names, [f'pattern_correlation_{x}' for x in range(idata.obsm['pattern_score'].shape[1])]] = corr_df.to_numpy()
        return idata
        
    def cell_transform(self, idata, adata, label=None):
        from scanpy.tools import rank_genes_groups
        import anndata
        adata = adata[adata.obs_names.isin(idata.uns['cell_meta'].index)]
        util.scored_spot_interface(idata)
        util.interaction_spot_interface(idata)
        adata.obsm['interaction_pattern'] = idata.uns['cell_pattern'].loc[adata.obs_names]
        adata.obsm['interaction_score'] = idata.uns['cell_score'].loc[adata.obs_names]
        print(f'Added key interaction_pattern, interaction_score in adata.obsm')
        
        if label is not None:
            adata_lri = anndata.AnnData(idata.uns['cell_score'])
            idata.uns['cell_meta'][label] = idata.uns['cell_meta'][label].astype(str).astype('category')
            small_clust = idata.uns['cell_meta'][label].value_counts()[idata.uns['cell_meta'][label].value_counts()==1].index.to_numpy()
            adata_lri.obs = idata.uns['cell_meta']
            adata_lri = adata_lri[~adata_lri.obs[label].isin(small_clust),:]
            rank_genes_groups(adata_lri, groupby=label)
            adata.uns['rank_interaction_score_groups'] = adata_lri.uns['rank_genes_groups']
            adata_pattern = anndata.AnnData(idata.uns['cell_pattern'])
            adata_pattern.obs = idata.uns['cell_meta']
            adata_pattern = adata_pattern[~adata_pattern.obs[label].isin(small_clust),:]
            rank_genes_groups(adata_pattern, groupby=label)
            adata.uns['rank_interaction_pattern_groups'] = adata_pattern.uns['rank_genes_groups']                                                          
            print(f'Added key rank_interaction_score_groups, rank_interaction_pattern_groups in adata.uns')   
        adata.obsm['interaction_pattern'] = adata.obsm['interaction_pattern'].to_numpy()                                                   
        adata.obsm['interaction_score'] = adata.obsm['interaction_score'].to_numpy()                                                   
        return adata, adata_lri, adata_pattern
    
    # def run_all(self, idata, adata, cluster_label, title, is_human):
    #     import matplotlib.pyplot as plt
    #     svi_df, svi_df_strict = spider.svi.combine_SVI(idata,threshold=0.01)
    #     self.svi.eva_SVI(idata, svi_df_strict)
    #     plt.savefig(f'../figures/{title}_metric.png')
    #     plt.close()
    #     merged_df,lri_pw_list,gene_lr_list,gene_pw_list = op.vis.svg_svi_relation(adata, idata, title=title, is_human=is_human)
    #     plt.savefig(f'../figures/{title}_intersect.png', dpi=600,bbox_inches='tight')
    #     idata.var[[f'pattern_correlation_{x}' for x in range(idata.obsm['pattern_score'].shape[1])]] = 0
    #     corr_df=pd.concat([idata[:,idata.var['is_svi']==1].to_df(),pd.DataFrame(idata.obsm['pattern_score'],index=idata.obs_names)],axis=1).corr().loc[idata[:,idata.var['is_svi']==1].var_names, range(idata.obsm['pattern_score'].shape[1])]
    #     idata.var.loc[idata[:,idata.var['is_svi']==1].var_names, [f'pattern_correlation_{x}' for x in range(idata.obsm['pattern_score'].shape[1])]] = corr_df.to_numpy()        
    #     op.vis.pattern_LRI(idata,show_SVI=10,spot_size=1)
    #     plt.tight_layout()
    #     plt.savefig(f'../figures/mouse_brain_st_celltrek_patterns.png', dpi=600,bbox_inches='tight')
    #     plt.close()
    







import pandas as pd
import numpy as np
import scanpy as sc
import networkx as nx
import magic
import scprep
from pathlib import Path
from sklearn.metrics import pairwise_distances
from scipy.spatial import Delaunay
PACKAGEDIR = Path(__file__).parent.absolute()
from sklearn.decomposition import PCA,NMF
import anndata 



# detect cci
def cci_spatalk(adata, work_dir, cluster_key, is_human, out_f):
    import os
    count_f = f'{work_dir}/adata_count.csv'
    meta_f = f'{work_dir}/adata_meta.csv'
    df = adata.to_df()
    df.index = "C"+df.index
    df.to_csv(count_f)
    meta = adata.obs[cluster_key].reset_index()
    meta[['x', 'y']] = adata.obsm['spatial']
    meta.columns = ['cell','celltype', 'x', 'y']
    meta.cell = "C"+meta.cell
    meta = meta[['cell','x', 'y', 'celltype']]
    if not pd.api.types.is_string_dtype(meta.celltype.dtype):
        meta.celltype = "T"+meta.celltype.astype('str')
    meta.celltype = meta.celltype.str.replace(' ', '_')
    meta.celltype = meta.celltype.str.replace('-', '_')
    meta.to_csv(meta_f)
    species = 'Human' if is_human else 'Mouse'
    os.system(str(f'/bin/bash -c "source /etc/profile;module load GCC/11.2.0 OpenMPI/4.1.1 R/4.2.0 Anaconda3/2022.05 R-bundle-Bioconductor/3.15-R-4.2.0;R -f ../spider/run_spatalk.R {count_f} {meta_f} {species} {out_f}"'))
    
# imputation
def impute_MAGIC(adata):
    magic_op = magic.MAGIC(n_jobs=5)
    inp = adata.to_df()
    inp = scprep.normalize.library_size_normalize(inp)
    inp = scprep.transform.sqrt(inp)
    outp = magic_op.fit_transform(inp)
    adata.X = outp

# idata
def idata_construct(score, pairs_meta, lr_df, lr_raw, adata):
    idata = anndata.AnnData(score)
    idata.obs_names = pairs_meta.index
    idata.var_names = lr_df.index
    idata.uns['lr_meta'] = lr_raw
    idata.obs = pairs_meta
    unique_cells = np.unique(idata.obs[['A', 'B']].to_numpy().flatten())
    cell_meta = adata.obs.loc[unique_cells]
    idata.uns['cell_meta'] = cell_meta
    # quality check
    sc.pp.calculate_qc_metrics(idata, inplace=True, percent_top=None)
    sc.pp.filter_genes(idata, min_cells=5)
    sc.pp.filter_cells(idata, min_genes=1)
    idata.obsm['spatial'] = idata.obs[['row', 'col']].to_numpy()
    print(f'Construct idata with {idata.shape[0]} interactions and {idata.shape[1]} LR pairs.')
    return idata

def score(adata, lr_df, pairs):
    exp_ref = adata.to_df()
    exp_ref = exp_ref.loc[:,~exp_ref.columns.duplicated()]
    l = lr_df['ligand'].to_numpy().flatten()
    r = lr_df['receptor'].to_numpy().flatten()
    sub_exp = exp_ref[np.concatenate((l, r))].to_numpy()
    sub_exp_rev = exp_ref[np.concatenate((r, l))].to_numpy()
    edge_exp_both = np.multiply(sub_exp[pairs[0]], sub_exp_rev[pairs[1]])
    score = np.sqrt(np.maximum(edge_exp_both[:, :int(len(l))], edge_exp_both[:, int(len(l)):]))
    return score

def subset(adata, lr_df):
    genes = adata.var_names.tolist()
    lr_df = lr_df[lr_df['ligand'].isin(genes) & lr_df['receptor'].isin(genes)]
    lr_df.index = lr_df['ligand'] + "_" + lr_df['receptor']
    l = lr_df['ligand'].to_numpy().flatten()
    r = lr_df['receptor'].to_numpy().flatten()
    unique_lr = np.unique(np.concatenate((l, r)))
    adata = adata[:, adata.var_names.isin(unique_lr)]
    sc.pp.filter_genes(adata, min_cells=1)
    sc.pp.filter_cells(adata, min_genes=1)
    sc.pp.normalize_total(adata, target_sum=1e4)
    genes = adata.var_names.tolist()
    lr_df = lr_df[lr_df['ligand'].isin(genes) & lr_df['receptor'].isin(genes)]
    return lr_df, adata

def find_pairs(adata, coord_type='grid', n_neighs=6):
    from squidpy.gr import spatial_neighbors
    from scipy.sparse import triu
    if coord_type == 'grid':
        spatial_neighbors(adata, coord_type=coord_type, n_neighs=n_neighs)
    else:
        spatial_neighbors(adata, coord_type=coord_type, delaunay=True, n_neighs=n_neighs)
    return np.transpose(triu(adata.obsp['spatial_connectivities']).nonzero()).T

# def meta(adata, position_keys, cluster_key, pairs):
def meta(adata, cluster_key, pairs):
    # get label
    pairs_meta = pd.DataFrame()
    pairs_meta['A'] = adata.obs_names[pairs[0]]
    pairs_meta['B'] = adata.obs_names[pairs[1]]
    # pairs_meta[['A_row', 'A_col']] = adata.obs[position_keys].iloc[pairs[0]].to_numpy()
    # pairs_meta[['B_row', 'B_col']] = adata.obs[position_keys].iloc[pairs[1]].to_numpy()
    pairs_meta[['A_row', 'A_col']] = adata.obsm['spatial'][pairs[0]]
    pairs_meta[['B_row', 'B_col']] =  adata.obsm['spatial'][pairs[1]]

    if cluster_key != '': 
        node_labels_text = adata.obs[cluster_key].to_numpy()
        pairs_meta['A_label'] = node_labels_text[pairs[0]].astype(str)
        pairs_meta['B_label'] = node_labels_text[pairs[1]].astype(str)
        node_labels = adata.obs[cluster_key].astype('category').cat.codes.to_numpy()
        pairs_meta['A_label_int'] = node_labels[pairs[0]]
        pairs_meta['B_label_int'] = node_labels[pairs[1]]
        pairs_meta['label_1'] = pairs_meta["A_label_int"].astype(str) + pairs_meta["B_label_int"].astype(str)
        pairs_meta['label_2'] = pairs_meta["B_label_int"].astype(str) + pairs_meta["A_label_int"].astype(str)
        # pairs_meta.to_csv('/home/lishiying/data6/01-interaction/check.csv')
        pairs_meta['label_int'] = pairs_meta[['label_1', 'label_2']].astype(int).max(axis=1).astype(str).astype('category')
        label_1 = pairs_meta['A_label'].astype(str) + '_' + pairs_meta['B_label'].astype(str).to_numpy()
        label_2 = pairs_meta['B_label'].astype(str) + '_' + pairs_meta['A_label'].astype(str).to_numpy()
        pick = pairs_meta[['label_1', 'label_2']].astype(int).idxmax(axis=1).to_numpy()
        text_label = [label_1[i] if x=='label_1' else label_2[i] for i,x in enumerate(pick)]
        pairs_meta['label'] = text_label
        pairs_meta['label'] = pairs_meta['label'].astype('category')

    pairs_meta.index = pairs_meta['A'] + "_" + pairs_meta['B']

    # get position  
    A_pos = pairs_meta[['A_row', 'A_col']].to_numpy(dtype=float)
    B_pos = pairs_meta[['B_row', 'B_col']].to_numpy(dtype=float)
    avg_pair_pos = (A_pos + B_pos) / 2
    pairs_meta[['row', 'col']] = avg_pair_pos
    pairs_meta['dist'] = np.linalg.norm(A_pos-B_pos, axis=1)
    return pairs_meta


# deplicated

def eigen_lr(idata, adata, lr_df, k, pairs, method='NMF'):
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)
    # adata.obs["total_counts"] = np.ravel(X.sum(axis=1))
    # sc.pp.normalize_total(to_reduce, target_sum=1e4)
    # print
    # sc.pp.log1p(to_reduce)
    # print(to_reduce)
    # sc.pp.regress_out(to_reduce, ['total_counts'])
    exp_l = adata[:, lr_df['ligand'].to_numpy().flatten()].to_df().to_numpy()
    exp_r = adata[:, lr_df['receptor'].to_numpy().flatten()].to_df().to_numpy()
    if method == 'NMF':
        model = NMF(n_components=k, init='random', random_state=0)
        W = model.fit_transform(np.vstack([exp_l, exp_r]))
    elif method == 'PCA':
        model = PCA(n_components=k, random_state=0, svd_solver='randomized') 
        W = model.fit_transform(np.vstack([exp_l, exp_r]))
    # adata.obsm[f'{method}_eigen_l'] = W[:len(to_reduce)]
    # adata.obsm[f'{method}_eigen_r'] = W[len(to_reduce):]
    idata.obsm[f'AL_eigen_{method}'] = W[:len(adata)][pairs[0]]
    idata.obsm[f'AR_eigen_{method}'] = W[len(adata):][pairs[0]]
    idata.obsm[f'BL_eigen_{method}'] = W[:len(adata)][pairs[1]]
    idata.obsm[f'BR_eigen_{method}'] = W[len(adata):][pairs[1]]


def load_lr_df(is_human):
    from importlib import resources
    with resources.path("spider.lrdb", "lrpairs.tsv") as pw_fn:
        lr_list = pd.read_csv(pw_fn, sep='\t', index_col=0)
    # lr_list = pd.read_csv(f'{PACKAGEDIR}/lr_pair_data/lrpairs.tsv', sep='\t', index_col=0)
    # lr_list['weight'] = 1
    # lr_list = lr_list[lr_list.species=='Human']
    
    if is_human:
        print('Using human LR pair dataset.')
        lr_list = lr_list[lr_list.species=='Human']
    else:
        print('Using mouse LR pair dataset.')
        lr_list = lr_list[lr_list.species=='Mouse']
    return lr_list


    # if is_human:
    #     print('Using human LR pair dataset.')
    #     lr = pd.read_csv(f'{PACKAGEDIR}/lr_pair_data/human(org_clean).txt', sep='\t', header=None)
    # else:
    #     print('Using mouse LR pair dataset.')
    #     lr = pd.read_csv(f'{PACKAGEDIR}/lr_pair_data/mouse.txt', sep='\t', header=None)

    lr_list = lr_list[['ligand','receptor', 'weight']]
    lr_list.columns = [0, 1, 2]
    # print(lr_list)
    return lr_list

def subset_old(adata, lr_df):
    # first subset
    genes = adata.var_names.tolist()

    lr_df = lr_df[lr_df['ligand'].isin(genes) & lr_df['receptor'].isin(genes)]
    # print(f'{len(lr_list)} out of {len(lr)} LR pairs are applicable.')

    # lr to df
    # lr_df = pd.DataFrame(lr_list)
    # lr_df.columns = ['ligands', 'receptors', 'weight']
    lr_df.index = lr_df['ligand'] + "_" + lr_df['receptor']

    # subset adata
    l = lr_df['ligand'].to_numpy().flatten()
    r = lr_df['receptor'].to_numpy().flatten()
    unique_lr = np.unique(np.concatenate((l, r)))

    # subset adata
    adata = adata[:, adata.var_names.isin(unique_lr)]
    sc.pp.filter_genes(adata, min_cells=1)
    sc.pp.filter_cells(adata, min_genes=2)
    # sc.pp.filter_genes(adata, min_cells=20)


    # processing adata
    # sc.pp.calculate_qc_metrics(adata, inplace=True)
    sc.pp.normalize_total(adata, target_sum=1e4)
    # sc.pp.log1p(adata)
    # sc.pp.highly_variable_genes(adata, min_mean=0.0125, max_mean=3, min_disp=0.5)
    # adata = adata[:, adata.var.highly_variable]
    # sc.pp.`regress_out`(adata, ['total_counts'])

    # second subset
    genes = adata.var_names.tolist()
    lr_df = lr_df[lr_df['ligand'].isin(genes) & lr_df['receptor'].isin(genes)]
    # if len(lr_df) < 10:
    #     raise Exception(f'Only {len(lr_df)} high-quality LR pairs out of {len(lr)} LR pairs are kept - Please check the quality of your data.')
    # else:
    #     print(f'{len(lr_df)} high-quality LR pairs out of {len(lr)} LR pairs are kept.')

    # second subset adata
    # l = lr_df['ligands'].to_numpy().flatten()
    # r = lr_df['receptors'].to_numpy().flatten()
    # unique_lr = np.unique(np.concatenate((l, r)))
    # adata = adata[:, unique_lr]
    # sc.pp.filter_genes(adata, min_cells=2)
    # sc.pp.filter_cells(adata, min_genes=1)
    return lr_df, adata

def nor_lr(lr_df):
    G = nx.Graph()
    edgeList = lr_df.to_numpy()
    for i in range(len(edgeList)):
        G.add_edge(edgeList[i][0], edgeList[i][1], weight=edgeList[i][2])
    A = nx.adjacency_matrix(G).A
    D = np.diag(1/np.sqrt(np.array(G.degree)[:, 1].astype(int)))
    nor_A = D @ A @ D 
    lr_df['weight'] = nor_A[np.nonzero(np.tril(nor_A, -1))]
    print(f'Weights of LR pairs are normalized.')

def impute_SMURF(adata):
    timer = TicToc()
    timer.tic()
    n_features = 10 if adata.shape[1] <= 100 else (20 if adata.shape[1] <= 500 else 50)
    operator = smurf.SMURF(n_features=n_features, estimate_only=True)
    adata.X = operator.smurf_impute(adata.to_df().T).T
    timer.toc('SMURF mputation finished in')



def impute_SAVER(adata):
    adata.to_df().to_csv(f'/home/lishiying/data3/mouse/{data[0]}/{data[1].split(".h5ad")[0]}_count.csv')
    
def find_pairs_v1(adata, position_keys, grid, n_neighs=6):
    dist = pairwise_distances(adata.obs[position_keys])
    if grid:
        if dist_offset == 0:
            dist = dist * (dist <= np.sort(dist, axis=1)[:,[20]]).astype(int)
        else:
            dist[dist > dist_offset] = 0
    else:
        print('non grid coordinate')
        tri = Delaunay(adata.obs[position_keys].to_numpy())
        indptr, indices = tri.vertex_neighbor_vertices
        adj = np.zeros((adata.shape[0], adata.shape[0]))
        for k in range(adata.shape[0]):
            neis = indices[indptr[k]:indptr[k+1]]
            for i in neis:
                adj[k][i] = 1
        adj += adj.T
        if dist_offset == 0:
            dist = dist * (dist <= np.quantile(dist, 0.5)).astype(int)
        else:
            dist[dist > np.quantile(dist, dist_offset)] = 0
        dist = dist * adj
    dist = np.tril(dist, -1)

    # weak_cells = np.where(~dist.any(axis=1))[0]

    # if len(weak_cells) != 0:
    #     adata_copy = adata_copy[~np.isin(np.arange(len(dist)), weak_cells)]
    #     dist = dist[~np.isin(np.arange(len(dist)), weak_cells)].T[~np.isin(np.arange(len(dist)), weak_cells)]
    #     print(f'dropped {len(weak_cells)} cells with no interaction with other cells.')
    
    return np.transpose(dist.nonzero()).T

# def normalize(adata):
#     timer = TicToc()
#     timer.tic()
#     librarySize = self.exp.sum(axis=0)
#     meanLibrarySize = librarySize.mean()
#     normalizedDataFrame = self.exp / (librarySize / meanLibrarySize)
#     self.exp = np.log(normalizedDataFrame+1)
#     timer.toc('Normalization finished in')

def count(adata, lr_df, edge_metric, pairs):
    exp_ref = adata.to_df()
    exp_ref = exp_ref.loc[:,~exp_ref.columns.duplicated()]
    l = lr_df['ligand'].to_numpy().flatten()
    r = lr_df['receptor'].to_numpy().flatten()

    # actual count
    AL = exp_ref.iloc[pairs[0]][l].to_numpy()
    AR = exp_ref.iloc[pairs[0]][r].to_numpy()
    BL = exp_ref.iloc[pairs[1]][l].to_numpy()
    BR = exp_ref.iloc[pairs[1]][r].to_numpy()
    # simplified expression
    sub_exp = exp_ref[np.concatenate((l, r))].to_numpy()
    sub_exp_rev = exp_ref[np.concatenate((r, l))].to_numpy()
    edge_exp_both = np.multiply(sub_exp[pairs[0]], sub_exp_rev[pairs[1]])
    if edge_metric == 'add':
        count = edge_exp_both[:, :int(len(l))] + edge_exp_both[:, int(len(l)):]
    elif edge_metric == 'max':
        count = np.maximum(edge_exp_both[:, :int(len(l))], edge_exp_both[:, int(len(l)):])
    direction = np.argmax((edge_exp_both[:, :int(len(l))], edge_exp_both[:, int(len(l)):]), 0)
    return AL, AR, BL, BR, direction, count





import scipy.optimize
import tensorflow as tf

from .util import concat_tensors, assign_concat


class MultiScipyOptimizer:
    def __init__(self, objective, variables):
        self.objective = objective
        self.variables = variables
        self._obj = self._wrap_func(objective, variables)

    def minimize(self, method="bfgs", **scipy_kwargs):
        res = scipy.optimize.minimize(
            self._obj,
            concat_tensors(self.variables).numpy(),
            method=method,
            jac=True,
            **scipy_kwargs,
        )
        assign_concat(res.x, self.variables)
        return res

    @classmethod
    def _wrap_func(cls, func, vars):
        def _objective(x):
            assign_concat(x, vars)
            with tf.GradientTape() as t:
                obj = func()
            grads = concat_tensors(
                t.gradient(obj, vars, unconnected_gradients=tf.UnconnectedGradients.ZERO)
            )
            return obj, grads

        _objective = tf.function(_objective)

        def _obj(x):
            loss, grad = _objective(x)
            return loss.numpy(), grad.numpy()

        return _obj



import pandas as pd
import numpy as np

from .preprocess import load_lr_df

def load_pathway(is_human=True):
    from importlib import resources
    with resources.path("spider.lrdb", "pathways.tsv") as pw_fn:
        pw = pd.read_csv(pw_fn, sep='\t', index_col=0)
    species = 'Human' if is_human else 'Mouse'
    pw = pw[pw.species==species].drop_duplicates()
    pw.index = pw.src + '_' + pw.dest
    return pw
    
def pathway_annotation(idata, is_human=True):
    pw = load_pathway(is_human)
    pw = pw.loc[pw.index.isin(idata.var_names)]
    df = pd.get_dummies(pw.pathway)
    df = df.groupby(df.index).sum()
    idata.varm['pathway'] = pd.DataFrame(0, columns=df.columns, index=idata.var_names)
    idata.varm['pathway'].loc[df.index] = df
    print(f'Added key pathway in idata.varm')

def pathway_annotation_list(LRI_list, is_human=True):
    pw = load_pathway(is_human)
    pw = pw.loc[pw.index.isin(LRI_list)]
    df = pd.get_dummies(pw.pathway)
    df = df.groupby(df.index).sum()
    return df

def pathway_prep(idata, is_human=True):
    lr_raw = load_lr_df(is_human=is_human)
    lr_raw.index = lr_raw.ligand + '_' + lr_raw.receptor
    pw = load_pathway(is_human)
    pw.index = pw.src + '_' + pw.dest
    pw = pw.loc[np.intersect1d(lr_raw.index, pw.index)]
    custom = {}
    for ptw in pw.pathway.unique():
        custom[ptw] = pw[pw.pathway == ptw].index.tolist()
    background = np.intersect1d(idata.var_names, pw.index).tolist()
    return custom, background

def pathway_prep_custom_background(background, is_human=True):
    lr_raw = load_lr_df(is_human=is_human)
    lr_raw.index = lr_raw.ligand + '_' + lr_raw.receptor
    pw = load_pathway(is_human)
    pw.index = pw.src + '_' + pw.dest
    pw = pw.loc[np.intersect1d(lr_raw.index, pw.index)]
    custom = {}
    for ptw in pw.pathway.unique():
        custom[ptw] = pw[pw.pathway == ptw].index.tolist()
    background = np.intersect1d(background, pw.index).tolist()
    return custom, background

def pattern_enrichment_edge(idata, is_human=True, groupby='label', subset=[], order=[], custom=None, background=None):
    import gseapy
    if not (custom and background):
        custom, background = pathway_prep(idata, is_human=is_human)
    var = idata.var
    if len(subset) != 0:
        var = var[var[groupby].isin(subset)]
    arr = []
    for i in var.sort_values(groupby)[groupby].unique():
        lri_list = var[var[groupby] ==i].index.to_numpy().tolist()
        arr.append([str(i), lri_list])
    dfs = []
    for i in range(len(arr)):
        try:
            enr_res = gseapy.enrichr(gene_list=arr[i][1],
                            gene_sets=custom,
                            background=background)
                            # cutoff = cutoff)
            enr_res.res2d[groupby] = arr[i][0]
            dfs.append(enr_res.res2d)
        except Exception as e:
            print(e)
            continue
    merged_df = pd.concat(dfs)
    if len(order)!=0:
        from string import ascii_lowercase
        rename_dict = {}
        count = 0
        for x in order:
            rename_dict[str(x)] = f'({ascii_lowercase[count]}) {x}'
            count += 1
        merged_df[f'ordered_{groupby}'] = merged_df[groupby]
        merged_df[f'ordered_{groupby}'] = merged_df[groupby].map(rename_dict)
    else:
        merged_df[f'ordered_{groupby}'] = merged_df[groupby]
    return merged_df, arr

def pattern_enrichment_node(idata, is_human=True, groupby='label', subset=[], order=[], custom=None, background=None, custom_pathwaydb=[]):
    import gseapy
    var = idata.var
    if len(subset) != 0:
        var = var[var[groupby].isin(subset)]
    arr = []
    for i in var.sort_values(groupby)[groupby].unique():
        lri_list = var[var[groupby] ==i].index.to_numpy().tolist()
        gene_list = np.unique(np.concatenate([x.split('_') for x in lri_list])).tolist()
        arr.append([str(i), gene_list])
    if len(custom_pathwaydb) == 0:
        pathway_db = ['KEGG_2021_Human' if is_human else 'KEGG_2019_Mouse']
    else:
        pathway_db = custom_pathwaydb
    organism = 'Human' if is_human else 'Mouse'
    dfs = []
    for i in range(len(arr)):
        try:
            enr_res = gseapy.enrichr(gene_list=arr[i][1],
                            organism=organism,
                            gene_sets=pathway_db)
            enr_res.res2d[groupby] = arr[i][0]
            dfs.append(enr_res.res2d)
        except Exception as e:
            print(e)
            continue
    merged_df = pd.concat(dfs)
    if len(order)!=0:
        from string import ascii_lowercase
        rename_dict = {}
        count = 0
        for x in order:
            rename_dict[str(x)] = f'({ascii_lowercase[count]}) {x}'
            count += 1
        merged_df[f'ordered_{groupby}'] = merged_df[groupby]
        merged_df[f'ordered_{groupby}'] = merged_df[groupby].map(rename_dict)
    else:
        merged_df[f'ordered_{groupby}'] = merged_df[groupby]
    return merged_df, arr
    
    
def enrichment(custom, background, histology_results, groupby='pattern', cutoff=0.01, top=None, order=[], group=[]):
    import gseapy
    dfs=[]
    histology_results = histology_results[histology_results.g.isin(np.intersect1d(histology_results.g, background))]
    arr = []
    if len(group) == 0:
        for i in histology_results.sort_values(groupby)[groupby].unique():
            lri_list = histology_results[histology_results[groupby] ==i].sort_values('membership', ascending=False)['g'].to_numpy()
            if top:
                lri_list = lri_list[:top]
            lri_list = lri_list.tolist()
            arr.append([str(i), lri_list])
    else:
        count = 0
        for x in order:
            index_m  = np.where(group == x)[0].tolist()
            lri_lists = []
            for i in index_m:
                lri_lists.append(histology_results.query('@groupby == @i').sort_values('membership', ascending=False)['g'].tolist())
            lri_list = np.concatenate(lri_lists)
            if top:
                lri_list = lri_list[:top]
            lri_list = lri_list.tolist()
            arr.append([f'Merged groupby {count}: {index_m}', lri_list])
            count += 1

    for i in range(len(arr)):
        try:
            enr_res = gseapy.enrichr(gene_list=arr[i][1],
                            gene_sets=custom,
                            background=background,
                            cutoff = cutoff)
            enr_res.res2d['group'] = arr[i][0]
            dfs.append(enr_res.res2d)
        except:
            print(i)
            continue
    merged_df = pd.concat(dfs)
    if len(order)!=0 and len(group) == 0:
        from string import ascii_lowercase
        rename_dict = {}
        count = 0
        for x in order:
            rename_dict[str(x)] = f'({ascii_lowercase[count]}) {x}'
            count += 1
        merged_df['ordered_group'] = merged_df.group
        merged_df['ordered_group'] = merged_df.group.map(rename_dict)
    else:
        merged_df['ordered_group'] = merged_df.group
    return merged_df, arr

def enrichment_interacrtion_list(idata, lri_list, is_human=True):
    import gseapy
    custom, background=pathway_prep(idata, is_human=is_human)
    enr_res = gseapy.enrichr(gene_list=lri_list,
                            gene_sets=custom,
                            background=background)
    return enr_res.results

def enrichment_interacrtion_gene_list(idata, lri_list, is_human=True,custom_pathwaydb=[]):
    import gseapy
    gene_list = np.unique(np.concatenate([x.split('_') for x in lri_list])).tolist()
    if len(custom_pathwaydb) == 0:
        pathway_db = ['KEGG_2021_Human' if is_human else 'KEGG_2019_Mouse']
    else:
        pathway_db = custom_pathwaydb
    organism = 'Human' if is_human else 'Mouse'
    enr_res = gseapy.enrichr(gene_list=gene_list,
                    organism=organism,
                    gene_sets=pathway_db)
    return enr_res.results

def enrichment_interacrtion_gene_df(idata, lri_df, groupby='label',is_human=True,custom_pathwaydb=[], order=[]):
    import gseapy
    arr = []
    for i in lri_df.sort_values(groupby)[groupby].unique():
        lri_list = lri_df[lri_df[groupby] ==i].index.to_numpy().tolist()
        gene_list = np.unique(np.concatenate([x.split('_') for x in lri_list])).tolist()
        arr.append([str(i), gene_list])
    if len(custom_pathwaydb) == 0:
        pathway_db = ['KEGG_2021_Human' if is_human else 'KEGG_2019_Mouse']
    else:
        pathway_db = custom_pathwaydb
    organism = 'Human' if is_human else 'Mouse'
    dfs = []
    for i in range(len(arr)):
        try:
            enr_res = gseapy.enrichr(gene_list=arr[i][1],
                            organism=organism,
                            gene_sets=pathway_db)
            enr_res.res2d[groupby] = arr[i][0]
            dfs.append(enr_res.res2d)
        except Exception as e:
            print(e)
            continue
    merged_df = pd.concat(dfs)
    if len(order)!=0:
        from string import ascii_lowercase
        rename_dict = {}
        count = 0
        for x in order:
            rename_dict[str(x)] = f'({ascii_lowercase[count]}) {x}'
            count += 1
        merged_df[f'ordered_{groupby}'] = merged_df[groupby]
        merged_df[f'ordered_{groupby}'] = merged_df[groupby].map(rename_dict)
    else:
        merged_df[f'ordered_{groupby}'] = merged_df[groupby]
    return merged_df, arr



def add_diff(idata, list1, list2, key_to_add):
    from sklearn.preprocessing import MinMaxScaler
    scaler = MinMaxScaler()
    exp1=scaler.fit_transform(np.sqrt(idata[:, idata.var_names.isin(list1)].to_df()).sum(axis=1).to_numpy().reshape(-1, 1)).flatten()
    exp2=scaler.fit_transform(np.sqrt(idata[:, idata.var_names.isin(list2)].to_df()).sum(axis=1).to_numpy().reshape(-1, 1)).flatten()
    idata.obs[key_to_add]=exp1-exp2
    print(f'Added key {key_to_add} in idata.uns')


from time import time
from typing import Optional, List, Tuple, Union
import warnings
from dataclasses import dataclass, field

from tqdm.auto import tqdm, trange
from anndata import AnnData
import numpy as np
import pandas as pd
import gpflow
from gpflow.utilities import to_default_float
import tensorflow as tf
import tensorflow_probability as tfp

from ._internal.svca import SVCA, SVCAInteractionScoreTest
from ._internal.optimizer import MultiScipyOptimizer
from ._internal.util import get_l_limits, bh_adjust, calc_sizefactors, dense_slice
from ._internal.distance_cache import DistanceCache


def test_spatial_interactions(
    adata: AnnData,
    layer: Optional[str] = None,
    spatial_key: str = "spatial",
    ard: bool = False,
    sizefactors: Optional[np.ndarray] = None,
    copy: bool = False,
) -> Tuple[pd.DataFrame, Union[AnnData, None]]:
    """
    Fit an SVCA model (Arnol, 2019) and test for spatial cell-cell interactions.

    In contrast to the original publication, which used a Gaussian approximation with a
    likelihood ratio test, this fits a Poisson GLMM and uses a score test.

    Args:
        adata: The annotated data matrix.
        layer: Name of the AnnData object layer to use. By default ``adata.X`` is used.
        spatial_key: Key in ``adata.obsm`` where the spatial coordinates are stored.
        ard: Whether to use automatic relevance determination for the kernel. This amounts to
            a separate lengthscale for each spatial dimension.
        sizefactors: Scaling factors for the observations. Defaults to total read counts.
        copy: Whether to return a copy of ``adata`` with results or write the results into ``adata``
            in-place.

    Returns:
        A tuple. The first element is a Pandas DataFrame with the test results, the second is ``None``
        if ``copy == False`` or an ``AnnData`` object.
    """
    if sizefactors is None:
        sizefactors = calc_sizefactors(adata)

    X = adata.obsm[spatial_key]
    l_min, l_max = get_l_limits(DistanceCache(X))
    lscales = l_min if not ard else [l_min] * X.shape[1]
    kernel = gpflow.kernels.SquaredExponential(lengthscales=lscales)
    kernel.lengthscales.transform = tfp.bijectors.Sigmoid(
        low=to_default_float(0.5 * l_min), high=to_default_float(2 * l_max)
    )
    gpflow.set_trainable(kernel.variance, False)

    results = []
    parameters = []
    test = SVCAInteractionScoreTest(
        dense_slice(adata.X if layer is None else adata.layers[layer]), X, sizefactors, kernel
    )

    params = gpflow.utilities.parameter_dict(test.kernel[0])
    sortedkeys = sorted(params.keys())
    dtype = np.dtype([(k, params[k].dtype.as_numpy_dtype) for k in sortedkeys])

    for i, g in enumerate(tqdm(adata.var_names)):
        t0 = time()
        res, _ = test(i, None)
        t = time() - t0
        results.append({"time": t, "pval": res.pval.numpy(), "gene": g})
        params = gpflow.utilities.read_values(test.kernel[0])
        parameters.append(tuple([params[k] for k in sortedkeys]))

    results = pd.DataFrame(results)
    results.loc[
        results.pval > 1, "pval"
    ] = 1  # this seems to be a bug in tensorflow_probability, survival_function should never be >1
    results["padj"] = bh_adjust(results.pval.to_numpy())

    if copy:
        adata = adata.copy()
        toreturn = adata
    else:
        toreturn = None
    adata.varm["svca"] = np.array(parameters, dtype=dtype)
    adata.obsm["svca_sizefactors"] = sizefactors
    adata.uns["svca_ard"] = ard
    adata.uns["svca_layer"] = layer

    return results, toreturn


def fit_spatial_interactions(
    adata: AnnData,
    layer: Optional[str] = None,
    genes: Optional[List[str]] = None,
    spatial_key: str = "spatial",
    ard: bool = False,
    sizefactors: Optional[np.ndarray] = None,
) -> pd.DataFrame:
    """
    Estimate magnitude of spatial cell-cell interactions using an SVCA model (ARnol, 2019).

    In contrast to the original publication, which used a Gaussian approximation, this fits
    a Poisson GLMM. This function is intendend to be used after :py:func:`test_spatial_interactions`
    to analyse the genes showing significant cell-cell interactions.

    Args:
        adata: The annotated data matrix.
        layer: Name of the AnnData object layer to use. By default ``adata.X`` is used.
        genes: List of genes to analyze. Defaults to all genes.
        spatial_key: Key in ``adata.obsm`` where the spatial coordinates are stored.
        ard: Whether to use automatic relevance determination for the kernel. This amounts to
            a separate lengthscale for each spatial dimension.
        sizefactors: Scaling factors for the observations. Defaults to total read counts.

    Returns: A Pandas DataFrame with the results.
    """
    if (
        "svca" not in adata.varm
        or "svca_sizefactors" not in adata.obsm
        or "svca_ard" not in adata.uns
        or "svca_layer" not in adata.uns
    ):
        warnings.warn("SVCA parameters not found in adata. Performing ab initio fitting.")
        if genes is None:
            genes = adata.var_names
        if sizefactors is None:
            sizefactors = calc_sizefactors(adata)
        trainable = True
    else:
        if genes is None:
            warnings.warn("No genes given. Fitting all genes.")
            genes = adata.var_names
        sizefactors = adata.obsm["svca_sizefactors"]
        ard = adata.uns["svca_ard"]
        trainable = False

    X = adata.obsm[spatial_key]
    l_min, l_max = get_l_limits(DistanceCache(X))
    lscales = l_min if not ard else [l_min] * X.shape[1]
    kernel = gpflow.kernels.SquaredExponential(lengthscales=lscales)
    kernel.lengthscales.transform = tfp.bijectors.Sigmoid(
        low=to_default_float(0.5 * l_min), high=to_default_float(2 * l_max)
    )
    gpflow.set_trainable(kernel.variance, False)

    model = SVCA(
        dense_slice(
            adata.X if adata.uns["svca_layer"] is None else adata.layers[adata.uns["svca_layer"]]
        ),
        X,
        sizefactors,
        kernel,
    )
    model.use_interactions(True)

    idx = np.argsort(adata.var_names)
    idx = idx[np.searchsorted(adata.var_names.to_numpy(), genes, sorter=idx)]
    if not trainable:
        param_names = adata.varm["svca"].dtype.names

    results = []
    for g, i in zip(tqdm(genes), idx):
        model.currentgene = i
        if not trainable:
            gpflow.utilities.multiple_assign(
                model.kernel, {n: v for n, v in zip(param_names, adata.varm["svca"][i])}
            )
        t0 = time()
        model.optimize()
        t = time() - t0
        fracvars = model.fraction_variance()._asdict()
        fracvars.update({"gene": g, "time": t})
        results.append(fracvars)
    return pd.DataFrame(results)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from .preprocess import load_lr_df

def viz_interface_pattern(idata, pos_key=['row', 'col'], label='',  histology=None):
    plt.figure(figsize=(15, 15))
    if label != '':
        plt.subplot(6, 5, 1)
        plt.scatter(idata.uns['cell_meta'][pos_key[0]], idata.uns['cell_meta'][pos_key[1]], c=idata.uns['cell_meta'][label].astype("category").cat.codes, s=1)
        plt.axis('equal')
        base = 2
    else:
        base = 1
    for i in range(idata.uns['cell_pattern'].shape[1]):
        plt.subplot(6, 5, i + base)
        plt.scatter(idata.uns['cell_meta'][pos_key[0]], idata.uns['cell_meta'][pos_key[1]], c=idata.uns['cell_pattern'][i], s=1)
        plt.axis('equal')
        if histology:
            plt.title('Pattern {} - {} LRIs'.format(i, histology.query('pattern == @i').shape[0] ))
        plt.colorbar(ticks=[])
        
def pattern_LRI(idata, label='', pos_key=[], obsm_key='pattern_score', show_SVI=0,spot_size=1):
    plt.figure(figsize=(30, 50))
    if label != '':
        plt.subplot(20, 11, 1)
        ax=sns.scatterplot(data=idata.uns['cell_meta'], x=pos_key[0], y=pos_key[1], hue=label, s=20, palette ='tab20')
        # ax=sns.scatterplot(data=idata.obs, x='row', y='col', hue=label, s=10, palette ='tab20')
        sns.move_legend(ax, "upper left", bbox_to_anchor=(1, 1.05),ncol=int(len(idata.uns['cell_meta'][label].unique())/8+1), title=None, frameon=False)
        plt.axis('equal')
        plt.axis('off')
        base = 9
    else:
        base = 1
    if show_SVI==0:
        for i in range(idata.obsm[obsm_key].shape[1]):
            plt.subplot(20, 11, i + base)
            im=plt.scatter(idata.obs['row'], idata.obs['col'], c=idata.obsm[obsm_key][:,i], s=spot_size, cmap='plasma')
            plt.axis('equal')
            plt.axis('off')
            plt.title('Pattern {} - {} LRIs'.format(i, np.sum(idata.var.label == i)) )
            plt.colorbar(im,fraction=0.046, pad=0.04)
    else:
        for i in range(idata.obsm[obsm_key].shape[1]):
            plt.subplot(20, 11, base)
            im=plt.scatter(idata.obs['row'], idata.obs['col'], c=idata.obsm[obsm_key][:,i], s=spot_size, cmap='plasma')
            plt.axis('equal')
            plt.axis('off')
            plt.title('Pattern {} - {} LRIs'.format(i, np.sum(idata.var.label == i)))
            plt.colorbar(im,fraction=0.046, pad=0.04)
            base += 1
            svis = idata.var[idata.var.label == i].sort_values(f'pattern_correlation_{i}', ascending=False)
            svi_to_plot = svis.index.to_numpy()[:show_SVI]
            memberships = svis[f'pattern_correlation_{i}'].to_numpy()[:show_SVI]
            for j in range(show_SVI):
                if j < len(svi_to_plot):
                    plt.subplot(20,11, base)
                    im=plt.scatter(idata.obs['row'], idata.obs['col'], c=idata.to_df()[svi_to_plot[j]], s=spot_size, cmap='plasma')
                    plt.axis('equal')
                    plt.axis('off')
                    plt.colorbar(im,fraction=0.046, pad=0.04)
                    plt.title(f'{svi_to_plot[j]}\n corr={"%.3f" % memberships[j]}')
                base += 1

def quiver_pattern(idata, label='', pos_key=[], obsm_key='pattern_score', traj_coords=[], show_SVI=0):
    plt.figure(figsize=(30, 50))
    if label != '':
        plt.subplot(20, 11, 1)
        ax=sns.scatterplot(data=idata.uns['cell_meta'], x=pos_key[0], y=pos_key[1], hue=label, s=10, palette ='tab20')
        # ax=sns.scatterplot(data=idata.obs, x='row', y='col', hue=label, s=10, palette ='tab20')
        sns.move_legend(ax, "upper left", bbox_to_anchor=(1, 1.05),ncol=int(len(idata.uns['cell_meta'][label].unique())/8+1), title=None, frameon=False)
        if len(traj_coords) != 0:
            for traj_cood in traj_coords:
                plt.quiver(traj_cood[0],traj_cood[2],traj_cood[1]-traj_cood[0],traj_cood[3]-traj_cood[2],color = 'red', angles='xy', scale_units='xy', scale = 1)
        plt.axis('equal')
        plt.axis('off')
        base = 9
    else:
        base = 1
    if show_SVI==0:
        for i in range(idata.obsm[obsm_key].shape[1]):
            plt.subplot(20, 11, i + base)
            im=plt.scatter(idata.obs['row'], idata.obs['col'], c=idata.obsm[obsm_key][:,i], s=1, cmap='plasma')
            if len(traj_coords) != 0:
                for traj_cood in traj_coords:
                    plt.quiver(traj_cood[0],traj_cood[2],traj_cood[1]-traj_cood[0],traj_cood[3]-traj_cood[2],color = 'red', angles='xy', scale_units='xy', scale = 1)
            plt.axis('equal')
            plt.axis('off')
            plt.title('Pattern {} - {} LRIs'.format(i, np.sum(idata.var.label == i)) )
            plt.colorbar(im,fraction=0.046, pad=0.04)
    else:
        for i in range(idata.obsm[obsm_key].shape[1]):
            plt.subplot(20, 11, base)
            im=plt.scatter(idata.obs['row'], idata.obs['col'], c=idata.obsm[obsm_key][:,i], s=1, cmap='plasma')
            if len(traj_coords) != 0:
                for traj_cood in traj_coords:
                    plt.quiver(traj_cood[0],traj_cood[2],traj_cood[1]-traj_cood[0],traj_cood[3]-traj_cood[2],color = 'red', angles='xy', scale_units='xy', scale = 1)
            plt.axis('equal')
            plt.axis('off')
            plt.title('Pattern {} - {} LRIs'.format(i, np.sum(idata.var.label == i)))
            plt.colorbar(im,fraction=0.046, pad=0.04)
            base += 1

            svis = idata.var[idata.var.label == i].sort_values(f'pattern_membership_{i}', ascending=False)
            svi_to_plot = svis.index.to_numpy()[:show_SVI]
            memberships = svis[f'pattern_membership_{i}'].to_numpy()[:show_SVI]
            for j in range(show_SVI):
                if j < len(svi_to_plot):
                    plt.subplot(20,11, base)
                    im=plt.scatter(idata.obs['row'], idata.obs['col'], c=idata.to_df()[svi_to_plot[j]], s=1, cmap='plasma')
                    if len(traj_coords) != 0:
                        for traj_cood in traj_coords:
                            plt.quiver(traj_cood[0],traj_cood[2],traj_cood[1]-traj_cood[0],traj_cood[3]-traj_cood[2],color = 'red', angles='xy', scale_units='xy', scale = 1)
                    plt.axis('equal')
                    plt.axis('off')
                    plt.colorbar(im,fraction=0.046, pad=0.04)
                    plt.title(f'{svi_to_plot[j]}')
                base += 1

def enrichment(df, x_key='ordered_group', figsize=None, size=None, save=None, cutoff=0.1, top_term=10):
    import gseapy
    df=df[df[x_key]!='-1']
    if not figsize:
        figsize = (3, int(top_term*len(df[x_key].unique())/10))
    if not size:
        size = int(top_term*len(df[x_key].unique())/20)
    ax = gseapy.dotplot(df,figsize=figsize, 
                    x=x_key, 
                    top_term=top_term,
                    column='P-value',
                    cutoff=cutoff,
                    cmap = plt.cm.plasma, 
                    xticklabels_rot=30,
                    size=size, 
                    ofname=save,
                    show_ring=True)
    if not save:
        # ax.set_xlabel("")
        plt.show()

def svg_svi_relation(adata, idata, title, is_human=True, top=200):
    from matplotlib_venn import venn3,venn3_circles
    import gseapy
    
    organism = 'Human' if is_human else 'Mouse'
    pathway_db = 'KEGG_2021_Human' if is_human else 'KEGG_2019_Mouse'
    
    pathway_dfs=[]

    print(organism, pathway_db)

    genelist = adata.uns['moranI'][:top].index.to_list()
    enr_res_gene = gseapy.enrichr(gene_list=genelist,
                            cutoff=0.01,
                            organism=organism,
                            gene_sets=pathway_db)
    gene_pw_list = enr_res_gene.res2d[enr_res_gene.res2d['P-value']<0.01].Term.to_numpy()
    enr_res_gene.res2d['group']='SVG'
    pathway_dfs.append(enr_res_gene.res2d[enr_res_gene.res2d['P-value']<0.01])
    
    
    human_lr = load_lr_df(is_human)
    l = human_lr['ligand'].to_numpy().flatten()
    r = human_lr['receptor'].to_numpy().flatten()
    unique_lr_human = np.unique(np.concatenate((l, r)))
    genelist_lr = adata.uns['moranI'].loc[np.intersect1d(unique_lr_human,adata.uns['moranI'].index)].sort_values('I',ascending=False)[:top].index.to_list()
    print(len(genelist_lr))
    if len(genelist_lr)!=0:
        enr_res_gene = gseapy.enrichr(gene_list=genelist_lr,
                                cutoff=0.01,
                                organism=organism,
                                gene_sets=pathway_db)
        gene_lr_list = enr_res_gene.res2d[enr_res_gene.res2d['P-value']<0.01].Term.to_numpy()
        enr_res_gene.res2d['group']='LR genes'
        pathway_dfs.append(enr_res_gene.res2d[enr_res_gene.res2d['P-value']<0.01])
    else:
        gene_lr_list=[]

    interaction_list = idata.uns['moranI'][:top].index.to_list()
    gene_list_lri = np.unique(np.concatenate([x.split('_') for x in interaction_list])).tolist()
    enr_res = gseapy.enrichr(gene_list=gene_list_lri,
                                cutoff=0.01,
                                organism=organism,
                                gene_sets=pathway_db)
    lri_pw_list = enr_res.res2d[enr_res.res2d['P-value']<0.01].Term.to_numpy()
    enr_res.res2d['group']='SVI genes'
    pathway_dfs.append(enr_res.res2d[enr_res.res2d['P-value']<0.01])

    fig = plt.figure(figsize=(12, 4))
    fig.suptitle(title, fontsize=15)
    plt.subplot(1, 3, 1)
    moran_df = adata.uns['moranI'][:top]
    moran_df['label'] = 'SVG'
    moran_df_lr = adata.uns['moranI'].loc[genelist_lr]
    moran_df_lr['label'] = 'LR genes'
    moran_df_svi = adata.uns['moranI'].loc[np.intersect1d(gene_list_lri,adata.uns['moranI'].index)]
    moran_df_svi['label'] = 'SVI genes'
    df = pd.concat([moran_df, moran_df_lr,moran_df_svi])
    print(moran_df.I.mean(),moran_df_lr.I.mean(), moran_df_svi.I.mean())
    sns.boxplot(data=df, x="label", y="I", order=["SVI genes", "LR genes",  "SVG",], palette={"SVI genes":"#098154","LR genes":  "#069af3",  "SVG": "#c72e29"})
    plt.title('Moran I')
    
    plt.subplot(1, 3, 2)
    g=venn3(subsets = [set(gene_list_lri),set(genelist_lr),set(genelist)], #绘图数据集
        set_labels = ('SVI genes', "LR genes",'SVG'), #设置组名
        set_colors=("#098154","#069af3","#c72e29"),#设置圈的颜色，中间颜色不能修改
        alpha=0.6,#透明度
        normalize_to=1.0,#venn图占据figure的比例，1.0为占满
       )
    plt.title('Gene Overlap')
    plt.subplot(1, 3, 3)
    g=venn3(subsets = [set(lri_pw_list),set(gene_lr_list),set(gene_pw_list)], #绘图数据集
        set_labels = ('SVI genes', "LR genes",'SVG'), #设置组名
        set_colors=("#098154","#069af3","#c72e29"),#设置圈的颜色，中间颜色不能修改
        alpha=0.6,#透明度
        normalize_to=1.0,#venn图占据figure的比例，1.0为占满
       )
    plt.title('Enriched Pathway Overlap')
    
    merged_df = pd.concat(pathway_dfs)

    return merged_df,lri_pw_list,gene_lr_list,gene_pw_list

def paga_compare(idata, save=None):
    import scanpy as sc
    sc.pl.paga_compare(idata, legend_fontsize=5, frameon=False, size=20, legend_fontoutline=2, show=False)
    if save:
        plt.savefig(save, bbox_inches="tight",dpi=300)
        
def transient_LRI(idata, label, save=None):
    import scvelo as scv
    idata.layers["spliced"] = idata.X
    idata.layers["unspliced"] = idata.X
    scv.tl.recover_dynamics(idata)
    top_genes = idata.var['fit_likelihood'].sort_values(ascending=False).index[:300]
    scv.pl.heatmap(idata, var_names=top_genes, tkey='dpt_pseudotime', n_convolve=30, col_color=label,sortby=label,save=save)
    
def plot_projection(idata, save=None):
    import scvelo as scv
    scv.pl.velocity_embedding_stream(
        idata, color="label", vkey="T_fwd_spatial", basis="spatial", legend_loc="right",
        recompute=False,V=idata.obsm['T_fwd_spatial'],smooth=1.5,
        save=save
    )


from time import time
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from typing import Optional, Union

import numpy as np
import pandas as pd
import gpflow
import tensorflow as tf

from .models import Model
from .sm_kernel import *
from .util import gower_factor


class Linear(gpflow.kernels.Linear):
    def K_novar(self, X, X2=None):
        if X2 is None:
            return tf.matmul(X, X, transpose_b=True)
        else:
            return tf.tensordot(X, X2, [-1, -1])


class SMPlusLinearKernel(gpflow.kernels.Sum):
    def __init__(self, sm_kernel):
        super().__init__([sm_kernel, Linear()])

    def K_novar(self, X, X2=None):
        return self._reduce([k.K_novar(X, X2) for k in self.kernels])

    @property
    def spectral_mixture(self):
        return self.kernels[0]

    @property
    def linear(self):
        return self.kernels[1]

    @staticmethod
    def _scaled_var(X, k):
        return gower_factor(k(X), k.variance)

    def scaled_variance(self, X):
        smvars = tf.convert_to_tensor([self._scaled_var(X, k) for k in self.kernels[0]])
        linvar = self._scaled_var(X, self.kernels[1])
        return tf.concat((smvars, tf.expand_dims(linvar, axis=0)), axis=0)


class GeneGPModel(metaclass=ABCMeta):
    @abstractmethod
    def freeze(self):
        pass

    @staticmethod
    def mixture_kernel(X, Y, ncomponents=5, ard=True, minvar=1e-3):
        range = tf.reduce_min(tf.reduce_max(X, axis=0) - tf.reduce_min(X, axis=0))
        dist = tf.sqrt(gpflow.utilities.ops.square_distance(X, None))
        dist = tf.linalg.set_diag(dist, tf.fill((X.shape[0],), tf.cast(np.inf, dist.dtype)))
        min_1nndist = tf.reduce_min(dist)

        datarange = tf.math.reduce_max(
            tf.math.reduce_max(X, axis=0) - tf.math.reduce_min(X, axis=0)
        )

        minperiod = 2 * min_1nndist
        varinit = min_1nndist + 0.5 * (range - min_1nndist)
        periodinit = minperiod + 0.5 * (range - minperiod)

        if ard:
            varinit = tf.repeat(varinit, X.shape[1])
            periodinit = tf.repeat(periodinit, X.shape[1])

        maxvar = 10 * tf.math.reduce_variance(Y)
        kernels = []
        for v in np.linspace(minvar, np.minimum(1, 0.9 * maxvar), ncomponents):
            k = Spectral(
                variance=gpflow.Parameter(
                    v,
                    transform=tfp.bijectors.Sigmoid(
                        low=gpflow.utilities.to_default_float(0),
                        high=gpflow.utilities.to_default_float(maxvar),
                    ),
                ),
                lengthscales=gpflow.Parameter(
                    varinit, transform=tfp.bijectors.Sigmoid(low=0.1 * min_1nndist, high=datarange)
                ),
                periods=gpflow.Parameter(
                    periodinit, transform=tfp.bijectors.Sigmoid(low=minperiod, high=2 * datarange)
                ),
            )
            kernels.append(k)
        kern = SpectralMixture(kernels)
        return SMPlusLinearKernel(kern)


class GPR(gpflow.models.GPR, GeneGPModel):
    def __init__(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        n_kernel_components: int = 5,
        ard: bool = True,
        minvar: float = 1e-3,
    ):
        kern = self.mixture_kernel(X, Y, n_kernel_components, ard, minvar)
        super().__init__(data=[X, Y], kernel=kern, mean_function=gpflow.mean_functions.Constant())

    def freeze(self):
        X = self.data[0]
        self.data = (None, self.data[1])
        frozen = gpflow.utilities.freeze(self)
        frozen.data = self.data = (X, self.data[1])
        return frozen


class SGPR(gpflow.models.SGPR, GeneGPModel):
    def __init__(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        inducing_variable: Union[np.ndarray, gpflow.inducing_variables.InducingPoints],
        n_kernel_components: int = 5,
        ard: bool = True,
        minvar: float = 1e-3,
    ):
        kern = self.mixture_kernel(X, Y, n_kernel_components, ard, minvar)
        super().__init__(
            data=[X, Y],
            kernel=kern,
            inducing_variable=inducing_variable,
            mean_function=gpflow.mean_functions.Constant(),
        )

    def freeze(self):
        X = self.data[0]
        self.data = (None, self.data[1])

        trainable_inducers = self.inducing_variable.Z.trainable
        if not trainable_inducers:
            Z = self.inducing_variable.Z
        frozen = gpflow.utilities.freeze(self)
        frozen.data = self.data = (X, self.data[1])
        if not trainable_inducers:
            frozen.inducing_variable.Z = Z
        return frozen

    def log_marginal_likelihood(self):
        return self.elbo()


@dataclass(frozen=True)
class VarPart:
    spectral_mixture: tf.Tensor
    linear: tf.Tensor
    noise: tf.Tensor


@dataclass(frozen=True)
class Variance:
    scaled_variance: VarPart
    fraction_variance: VarPart
    var_fraction_variance: VarPart


class GeneGP(Model):
    def __init__(self, model: GeneGPModel, minimize_fun, *args, **kwargs):
        self.model = model

        self._frozen = False
        self._variancevars = list(self.model.kernel.parameters)
        self._variancevars.append(self.model.likelihood.variance)

        self._trainable_variance_idx = []
        offset = 0
        variancevars = set(self._variancevars)
        for v in self.model.parameters:
            if v in variancevars:
                self._trainable_variance_idx.extend([offset + int(i) for i in range(tf.size(v))])
                offset += int(tf.size(v))

        self.__invHess = None
        self.model.likelihood.variance.assign(tf.math.reduce_variance(self.model.data[1]))

        t0 = time()
        self._optimize(minimize_fun, *args, **kwargs)
        self._freeze()
        t = time() - t0

        self._time = t

    @property
    def kernel(self):
        return self.model.kernel

    @property
    def K(self):
        return self.model.kernel.K_novar(self.model.data[0])

    @property
    def y(self):
        return tf.squeeze(self.model.data[1]).numpy()

    def predict_mean(self, X=None):
        if X is None:
            X = self.model.data[0]
        return self.model.predict_f(X)[0]

    def plot_power_spectrum(self, xlim: float = None, ylim: float = None, **kwargs):
        return self.model.kernel.spectral_mixture.plot_power_spectrum(xlim, ylim, **kwargs)

    @property
    def time(self):
        return self._time

    @staticmethod
    def _concat_tensors(tens):
        return tf.concat([tf.reshape(t, (-1,)) for t in tens], axis=0)

    @property
    def _invHess(self):
        if self.__invHess is None:
            # tf.hessians doesn't work yet (https://github.com/tensorflow/tensorflow/issues/29781)
            # and tape.jacobian() doesn't like lengthscale and period parameters for some reason
            # (it aborts with AttributeError: Tensor.graph is meaningless when eager execution is enabled.).
            # So we need to do this the hard way
            with tf.GradientTape(persistent=True) as tape:
                y = self.model.log_marginal_likelihood()
                tape.watch(y)
                grad = self._concat_tensors(tape.gradient(y, self.model.trainable_variables))
                grads = tf.split(
                    grad, tf.ones((tf.size(grad),), dtype=tf.int32)
                )  # this is necessary to be able to get the gradient of each entry
            hess = tf.stack(
                [
                    self._concat_tensors(
                        tape.gradient(
                            g,
                            self.model.trainable_variables,
                            unconnected_gradients=tf.UnconnectedGradients.ZERO,
                        )
                    )
                    for g in grads
                ]
            )
            self._invHess = tf.linalg.inv(hess)
        return self.__invHess

    @_invHess.setter
    def _invHess(self, invhess):
        x, y = tf.meshgrid(self._trainable_variance_idx, self._trainable_variance_idx)
        invhess = tf.reshape(
            tf.gather_nd(invhess, tf.stack([tf.reshape(x, (-1,)), tf.reshape(y, (-1,))], axis=1)),
            (len(self._trainable_variance_idx), len(self._trainable_variance_idx)),
        )
        self.__invHess = invhess

    def _optimize(self, minimize_fun, *args, **kwargs):
        res = minimize_fun(
            lambda: -self.model.log_marginal_likelihood(),
            self.model.trainable_variables,
            *args,
            **kwargs,
        )
        if isinstance(res, dict) and "hess_inv" in res:
            self._invHess = gpflow.utilities.to_default_float(res["hess_inv"])
        elif hasattr(res, "hess_inv"):
            self._invHess = gpflow.utilities.to_default_float(res.hess_inv)

    def _freeze(self):
        if self._frozen:
            return
        # this code calculates the variance of the fraction of spatial variance estimate
        # We use the negative of the Hessian of the marginal log-likelihood as observed Fisher observation, the inverse of which is the
        # asymptotic covariance matrix of the estimate. We then use the Delta method to get the asymptotic variance of FSV.
        # TODO: I'm not quite sure if this is valid for the case of free inducing points, since these are variational parameters
        with tf.GradientTape() as t:
            variances = self._concat_tensors(
                [
                    self.model.kernel.scaled_variance(self.model.data[0]),
                    tf.expand_dims(self.model.likelihood.variance, axis=0),
                ]
            )
            totalvar = tf.reduce_sum(variances)
            variance_fractions = variances / totalvar

        grads = t.jacobian(
            variance_fractions, [v.unconstrained_variable for v in self._variancevars]
        )
        grads = tf.concat([tf.expand_dims(g, -1) if g.ndim < 2 else g for g in grads], axis=1)
        errors = tf.reduce_sum((grads @ self._invHess) * grads, axis=1)

        variances = VarPart(variances[0:-2], variances[-2], variances[-1])
        variance_fractions = VarPart(
            variance_fractions[0:-2], variance_fractions[-2], variance_fractions[-1]
        )
        errors = VarPart(errors[0:-2], errors[-2], errors[-1])

        self.variances = Variance(variances, variance_fractions, errors)

        self.model = self.model.freeze()
        self._frozen = True


class DataSetResults(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value: GeneGP):
        if not isinstance(value, GeneGP):
            raise TypeError("value must be a GeneGP object")
        super().__setitem__(key, value)

    def to_df(self, modelcol: str = "model"):
        df = defaultdict(lambda: [])
        for gene, res in self.items():
            df["gene"].append(gene)
            variances = res.variances
            df["FSV"].append((1 - variances.fraction_variance.noise).numpy())
            df["s2_FSV"].append(variances.var_fraction_variance.noise.numpy())

            for i, k in enumerate(res.kernel.spectral_mixture.kernels):
                df["sm_variance_%i" % i].append(k.variance.numpy())
                df["sm_lengthscale_%i" % i].append(k.lengthscales.numpy())
                df["sm_period_%i" % i].append(k.periods.numpy())
            df["linear_variance"].append(res.kernel.linear.variance.numpy())
            df["noise_variance"].append(res.model.likelihood.variance.numpy())
            df["time"].append(res.time)
            df[modelcol].append(res)
        df = pd.DataFrame(df)
        return df
